import csv
import gzip
import io
import logging
import os
import struct
import zipfile

_sqlite_keywords = {
    "ABORT",
    "ACTION",
    "ADD",
    "AFTER",
    "ALL",
    "ALTER",
    "ANALYZE",
    "AND",
    "AS",
    "ASC",
    "ATTACH",
    "AUTOINCREMENT",
    "BEFORE",
    "BEGIN",
    "BETWEEN",
    "BY",
    "CASCADE",
    "CASE",
    "CAST",
    "CHECK",
    "COLLATE",
    "COLUMN",
    "COMMIT",
    "CONFLICT",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "DATABASE",
    "DEFAULT",
    "DEFERRABLE",
    "DEFERRED",
    "DELETE",
    "DESC",
    "DETACH",
    "DISTINCT",
    "DROP",
    "EACH",
    "ELSE",
    "END",
    "ESCAPE",
    "EXCEPT",
    "EXCLUSIVE",
    "EXISTS",
    "EXPLAIN",
    "FAIL",
    "FOR",
    "FOREIGN",
    "FROM",
    "FULL",
    "GLOB",
    "GROUP",
    "HAVING",
    "IF",
    "IGNORE",
    "IMMEDIATE",
    "IN",
    "INDEX",
    "INDEXED",
    "INITIALLY",
    "INNER",
    "INSERT",
    "INSTEAD",
    "INTERSECT",
    "INTO",
    "IS",
    "ISNULL",
    "JOIN",
    "KEY",
    "LEFT",
    "LIKE",
    "LIMIT",
    "MATCH",
    "NATURAL",
    "NO",
    "NOT",
    "NOTNULL",
    "NULL",
    "OF",
    "OFFSET",
    "ON",
    "OR",
    "ORDER",
    "OUTER",
    "PLAN",
    "PRAGMA",
    "PRIMARY",
    "QUERY",
    "RAISE",
    "REFERENCES",
    "REGEXP",
    "REINDEX",
    "RELEASE",
    "RENAME",
    "REPLACE",
    "RESTRICT",
    "RIGHT",
    "ROLLBACK",
    "ROW",
    "SAVEPOINT",
    "SELECT",
    "SET",
    "TABLE",
    "TEMP",
    "TEMPORARY",
    "THEN",
    "TO",
    "TRANSACTION",
    "TRIGGER",
    "UNION",
    "UNIQUE",
    "UPDATE",
    "USING",
    "VACUUM",
    "VALUES",
    "VIEW",
    "VIRTUAL",
    "WHEN",
    "WHERE",
}


class SmartFileReader:
    def __init__(self, file, *args, **kwargs):
        if file[-3:] == ".gz":
            with open(file, "rb") as f:
                f.seek(-4, 2)
                self._filesize = struct.unpack("I", f.read(4))[0]
            self.file = gzip.open(file, "rt", *args, **kwargs)
        elif file[-4:] == ".zip":
            zf = zipfile.ZipFile(file, "r")
            zf_info = zf.infolist()
            if len(zf_info) != 1:
                raise TypeError(
                    "zip archive files must contain a single member file for SmartFileReader"
                )
            zf_info = zf_info[0]
            self.file = zf.open(zf_info.filename, "r", *args, **kwargs)
            self._filesize = zf_info.file_size
        else:
            self.file = open(file, *args, **kwargs)
            self._filesize = os.fstat(self.file.fileno()).st_size

    def __getattr__(self, name):
        return getattr(self.file, name)

    def __setattr__(self, name, value):
        if name in ["file", "percentread", "_filesize"]:
            return object.__setattr__(self, name, value)
        return setattr(self.file, name, value)

    def __delattr__(self, name):
        return delattr(self.file, name)

    def percentread(self):
        try:
            return float(self.file.tell()) / float(self._filesize) * 100
        except io.UnsupportedOperation:
            try:
                return 1.0 - (float(self.file._left) / float(self._filesize) * 100)
            except Exception:
                return 0

    def __iter__(self):
        return self.file.__iter__()

    def bytesread(self):
        try:
            b = float(self.file.tell())
        except Exception:
            return "error in bytesread"
        labels = ["B", "KB", "MB", "GB", "TB"]
        scale = 0
        while scale < 4 and b > 1024:
            b /= 1024
            scale += 1
        return f"{b:.2f}{labels[scale]}"

    def progress(self):
        pct = self.percentread()
        if pct > 0 and pct < 100:
            return f"{pct}%"
        else:
            return self.bytesread()


def interpret_header_name(h):
    eL = logging.getLogger("l4util")
    if h.lower() == "case" or h[1:].lower() == "case":
        eL.warning("'case' is a reserved keyword, changing column header to 'caseid'")
        h = h + "id"
    if h.lower() == "group" or h[1:].lower() == "group":
        eL.warning("'group' is a reserved keyword, changing column header to 'groupid'")
        h = h + "id"
    if h.upper() in _sqlite_keywords:
        eL.warning(f"'{h}' is a reserved keyword, changing column header to '{h}_'")
        h = h + "_"
    if "." in h:
        h = h.replace(".", "_")
        eL.warning(
            f"dots are not allowed in column headers, changing column header to '{h}'"
        )
    if len(h) == 0:
        eL.error("zero length header")
        return " "
    if h[0] == "@":
        return h[1:] + " TEXT"
    elif h[0] == "#":
        return h[1:] + " INTEGER"
    return h + " NUMERIC"


def prepare_import_headers(rawfilename, headers=None):
    """
    Identify the file type and read the column headers from the first row.

    For each column, data type is presumed to be FLOAT unless the column name is
    prepended with '@' or '#' which makes the column format TEXT or INT respectively.
    Since data is stored in SQLite, format is just a suggestion anyhow.

    Input:
      rawfile: the filename (absolute or relative path) to the text file with data.
    Output:
      headers: a series of 2-tuples identifying column names and data types
      r: the csv reader object, primed to begin reading at the second line of the file
    """
    eL = logging.getLogger("l4util")
    if isinstance(rawfilename, str):
        raw = SmartFileReader(rawfilename)
    else:
        raw = rawfilename
    sample = raw.read(28192)
    raw.seek(0)
    if isinstance(sample, bytes):
        sample = sample.decode("UTF-8")
    try:
        dialect = csv.Sniffer().sniff(sample)
    except csv.Error:
        sample = raw.read(512000)
        raw.seek(0)
        if isinstance(sample, bytes):
            sample = sample.decode("UTF-8")
        dialect = csv.Sniffer().sniff(sample)
    r = csv.reader(raw, dialect)
    eL.debug("DIALECT = %s", str(dialect))
    eL.debug("TYPE OF READER = %s", str(type(r)))
    if headers is None:
        try:
            headers = r.next()
        except AttributeError:
            headers = next(r)  ##
    # fix header items for type
    for h in range(len(headers)):
        headers[h] = interpret_header_name(headers[h])
    return headers, r, raw


def prepare_import_headers_plain(rawfilename, headers=None):
    """
    Identify the file type and read the column headers from the first row.

    For each column, data type is presumed to be FLOAT unless the column name is
    prepended with '@' or '#' which makes the column format TEXT or INT respectively.
    Since data is stored in SQLite, format is just a suggestion anyhow.

    Input:
      rawfile: the filename (absolute or relative path) to the text file with data.
    Output:
      headers: a series of 2-tuples identifying column names and data types
      r: the csv reader object, primed to begin reading at the second line of the file
    """
    eL = logging.getLogger("l4util")
    if isinstance(rawfilename, str):
        raw = SmartFileReader(rawfilename)
    else:
        raw = rawfilename
    sample = raw.read(28192)
    raw.seek(0)
    if isinstance(sample, bytes):
        sample = sample.decode("UTF-8")
    try:
        dialect = csv.Sniffer().sniff(sample)
    except csv.Error:
        sample = raw.read(512000)
        raw.seek(0)
        if isinstance(sample, bytes):
            sample = sample.decode("UTF-8")
        dialect = csv.Sniffer().sniff(sample)
    r = csv.reader(raw, dialect)
    eL.debug("DIALECT = %s", str(dialect))
    eL.debug("TYPE OF READER = %s", str(type(r)))
    if headers is None:
        try:
            headers = r.next()
        except AttributeError:
            headers = next(r)  ##
    return headers, r, raw



"""
Base model class and related functions for a discrete choice modeling framework.

The main class in this module is `BaseModel`, which serves as the base class for
all models in the framework. It contains a `ParameterBucket` for storing parameters,
a `DictOfLinearFunction` for computing utility from idco data, and a placeholder
for the model subtype.

The `BaseModel` class is designed to be subclassed by other model classes that
implement specific types of discrete choice models.
"""

import base64
import logging
import pathlib
import uuid
import warnings

import numpy as np
import pandas as pd
import xarray as xr

from .._optional import jax
from ..dataset import DataTree
from ..exceptions import MissingDataError
from .constraints import ParametricConstraintList
from .linear import DictOfAlts, DictOfLinearFunction, LinearFunction
from .mixtures import MixtureList
from .param_core import ParameterBucket
from .single_parameter import SingleParameter
from .tree import NestingTree

logger = logging.getLogger("larch.model")

MANGLE_DATA = 0x1
MANGLE_STRUCTURE = 0x2


def _unique_ident():
    j = base64.b32encode(uuid.uuid4().bytes)[:25].decode()
    return f"{j[:5]}-{j[5:10]}-{j[10:15]}-{j[15:20]}-{j[20:]}"


class BaseModel:
    """Base class for discrete choice models."""

    _parameter_bucket = ParameterBucket()
    _model_subtype = None

    utility_co = DictOfLinearFunction()
    """DictOfLinearFunction : The portion of the utility function computed from |idco| data.

    The keys of this mapping are alternative codes for the applicable elemental
    alteratives, and the values are linear functions to compute for the indicated
    alternative.  Each alternative that has any idco utility components must have
    a unique linear function given.

    Examples
    --------

    >>> from larch import Model, P, X
    >>> m = Model()
    >>> m.utility_co = {
    ... 	1: P.ParamA1 * X.DataA,
    ... 	2: P.ParamA2 * X.DataA + P.ParamASC2,
    ... }
    >>> print(m.utility_co)
    DictOfLinearFunction({1: P.ParamA1 * X.DataA, 2: P.ParamA2 * X.DataA + P.ParamASC2})
    """

    utility_ca = LinearFunction()
    """LinearFunction : The portion of the utility function computed from |idca| data.

    Data expressions in this utility function can actually reference both |idca|
    and |idco| format variables. Except in unusual model designs, every complete data
    expression should have at least one |idca| component.

    Examples
    --------

    >>> from larch import Model, P, X
    >>> m = Model()
    >>> m.utility_ca = P.Param1 * X.Data1 + P.Param2 * X.Data2
    >>> print(m.utility_ca)
    P.Param1 * X.Data1 + P.Param2 * X.Data2
    >>> m.utility_ca += P.Param3 * X.Data3 / X.DataCO4
    >>> print(m.utility_ca)
    P.Param1 * X.Data1 + P.Param2 * X.Data2 + P.Param3 * X('Data3/DataCO4')
    """

    quantity_ca = LinearFunction()
    """LinearFunction : The portion of the quantity function computed from |idca| data.

    Data expressions in this utility function can actually reference both |idca|
    and |idco| format variables. Except in unusual model designs, every complete data
    expression should have at least one |idca| component.

    Note that for the quantity function, the actual computed linear function
    uses the exponential of the parameter value(s), not the raw values. Thus,
    if the quantity function is given as `P.Param1 * X.Data1 + P.Param2 * X.Data2`,
    the computed values will actually be `exp(P.Param1) * X.Data1 + exp(P.Param2) * X.Data2`.
    This transformation ensures that the outcome from the quantity function is
    always positive, so long as at all of the data terms in the function are
    positive.  The `LinearFunction` class itself is not intrinsically aware
    of this implementation detail, but the `Model.utility_functions()` method is,
    and will render the complete utility function in a mathematically correct form.

    Examples
    --------

    >>> from larch import Model, P, X
    >>> m = Model()
    >>> m.quantity_ca = P.Param1 * X.Data1 + P.Param2 * X.Data2
    >>> print(m.quantity_ca)
    P.Param1 * X.Data1 + P.Param2 * X.Data2
    >>> m.quantity_ca += P.Param3 * X.Data3 / X.DataCO4
    >>> print(m.quantity_ca)
    P.Param1 * X.Data1 + P.Param2 * X.Data2 + P.Param3 * X('Data3/DataCO4')
    """

    _graph = NestingTree()

    constraints = ParametricConstraintList()

    _availability_co_vars = DictOfAlts()
    _choice_co_vars = DictOfAlts()

    mixtures = MixtureList()

    def __init__(
        self,
        *,
        title=None,
        datatree=None,
        compute_engine=None,
        submodels=None,
        named_submodels=None,
        use_streaming=False,
        cache_dir=None,
    ):
        if not hasattr(self, "_ident"):
            self._ident = _unique_ident()
        self._mangled = 0x3
        self._datatree = None
        self.title = title
        self.rename_parameters = {}
        self._compute_engine = compute_engine
        self._use_streaming = use_streaming

        if submodels is None:
            submodels = {}
        for k, m in submodels.items():
            m.ident = k
        if named_submodels is None:
            named_submodels = {}
        for k, m in named_submodels.items():
            m.ident = k
        bucket = ParameterBucket(submodels, **named_submodels)
        self._parameter_bucket = bucket
        self.datatree = datatree
        if cache_dir is not None:
            cache_dir = pathlib.Path(cache_dir)
            cache_dir.mkdir(exist_ok=True)
            self.datatree.cache_dir = cache_dir
        self._cached_loglike_best = None
        self._cached_loglike_null = None
        self._most_recent_estimation_result = None

        self._choice_ca_var = None
        self._choice_co_code = None
        self._choice_co_vars = None
        self._choice_any = None

        self._weight_co_var = None

        self._availability_ca_var = None
        self._availability_co_vars = None
        self._availability_any = True

        self.ordering = None

    @property
    def ident(self):
        """Getter method for the ident property."""
        return self._ident

    @ident.setter
    def ident(self, x):
        try:
            old_ident = self._ident
        except AttributeError:
            pass
        else:
            self._parameter_bucket.rename_model(old_ident, x)
        self._ident = x

    @property
    def compute_engine(self):
        return self._compute_engine

    @compute_engine.setter
    def compute_engine(self, engine):
        if engine not in {"numba", "jax", None}:
            raise ValueError("invalid compute engine")
        self._compute_engine = engine
        if self._compute_engine == "jax" and not jax:
            warnings.warn(
                "jax is not installed, falling back to numba",
                stacklevel=2,
            )
            self._compute_engine = "numba"
        if self._compute_engine == "jax" and self.use_streaming:
            warnings.warn(
                "setting use_streaming to False, jax is not yet compatible",
                stacklevel=2,
            )
            self.use_streaming = False

    @property
    def use_streaming(self):
        return self._use_streaming

    @use_streaming.setter
    def use_streaming(self, should):
        should = bool(should)
        if should and self.compute_engine != "numba":
            raise ValueError(
                "streaming is currently only compatible with the numba compute engine"
            )
        self._use_streaming = should

    @property
    def most_recent_estimation_result(self):
        """A copy of the result dict from most recent likelihood maximization."""
        return self._most_recent_estimation_result.copy()

    @property
    def possible_overspecification(self):
        """Not yet implemented."""
        return None

    @property
    def datatree(self):
        """DataTree : A source for data for the model."""
        return self._datatree

    @datatree.setter
    def datatree(self, tree):
        self.swap_datatree(tree, True)
        # from ..dataset import DataTree
        #
        # if tree is self.datatree:
        #     return
        # if isinstance(tree, DataTree) or tree is None:
        #     self._datatree = tree
        #     self.mangle(structure=False)
        # elif isinstance(tree, xr.Dataset):
        #     self._datatree = tree.dc.as_tree()
        #     self.mangle(structure=False)
        # else:
        #     try:
        #         self._datatree = DataTree(main=xr.Dataset.construct(tree))
        #     except Exception as err:
        #         raise TypeError(f"datatree must be DataTree not {type(tree)}") from err
        #     else:
        #         self.mangle(structure=False)

    def swap_datatree(self, tree: DataTree | xr.Dataset, should_mangle=False) -> None:
        """
        Swap the current datatree with a new datatree.

        Parameters
        ----------
        tree : DataTree or xr.Dataset or None
            The new datatree to be swapped. If None, the datatree will be set to None.
        should_mangle : bool, optional
            A boolean indicating whether to mangle the structure after swapping the datatree.
            Default is False.

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If the `tree` parameter is not of type DataTree or xr.Dataset.

        Notes
        -----
        If `tree` is a xarray.Dataset, it will be converted to a DataTree using
        the `as_tree` method of the Dataset.

        If `should_mangle` is True, the structure of the datatree will be mangled
        after swapping.

        Examples
        --------
        >>> model = BaseModel()
        >>> tree = DataTree()
        >>> model.swap_datatree(tree, should_mangle=True)
        """
        from ..dataset import DataTree

        if (
            tree is not None
            and hasattr(tree, "relationships_are_digitized")
            and (not tree.relationships_are_digitized)
            and self.use_streaming
        ):
            tree = tree.digitize_relationships()
        if tree is self.datatree:
            return
        if isinstance(tree, DataTree) or tree is None:
            self._datatree = tree
            if should_mangle:
                self.mangle(structure=False)
        elif isinstance(tree, xr.Dataset):
            self._datatree = tree.dc.as_tree()
            if should_mangle:
                self.mangle(structure=False)
        else:
            try:
                self._datatree = DataTree(main=xr.Dataset.construct(tree))
            except Exception as err:
                raise TypeError(f"datatree must be DataTree not {type(tree)}") from err
            else:
                if should_mangle:
                    self.mangle(structure=False)

    @property
    def parameters(self):
        self.unmangle()
        return self._parameter_bucket.parameters

    def update_parameters(self, x):
        self.unmangle()
        return self._parameter_bucket.update_parameters(x)

    def add_parameter_array(self, name, values):
        return self._parameter_bucket.add_array(name, values)

    @property
    def n_params(self):
        self.unmangle()
        return self._parameter_bucket.n_params

    @property
    def pvals(self):
        self.unmangle()
        return self._parameter_bucket.pvals

    @pvals.setter
    def pvals(self, x):
        self.unmangle()
        self._parameter_bucket.pvals = x

    @property
    def pnames(self):
        self.unmangle()
        return self._parameter_bucket.pnames

    @property
    def pholdfast(self):
        self.unmangle()
        return self._parameter_bucket.pholdfast

    @pholdfast.setter
    def pholdfast(self, x):
        self.unmangle()
        self._parameter_bucket.pholdfast = x

    @property
    def pnullvals(self):
        self.unmangle()
        return self._parameter_bucket.pnullvals

    @pnullvals.setter
    def pnullvals(self, x):
        self.unmangle()
        self._parameter_bucket.pnullvals = x

    @property
    def pmaximum(self):
        self.unmangle()
        return self._parameter_bucket.pmaximum

    @pmaximum.setter
    def pmaximum(self, x):
        self.unmangle()
        self._parameter_bucket.pmaximum = x

    @property
    def pminimum(self):
        self.unmangle()
        return self._parameter_bucket.pminimum

    @pminimum.setter
    def pminimum(self, x):
        self.unmangle()
        self._parameter_bucket.pminimum = x

    @property
    def pbounds(self):
        """scipy.optimize.Bounds : A copy of the current min-max bounds of the parameters."""
        self.unmangle()
        from scipy.optimize import Bounds

        return Bounds(
            self._parameter_bucket.pminimum,
            self._parameter_bucket.pmaximum,
        )

    @property
    def pstderr(self):
        self.unmangle()
        return self._parameter_bucket.pstderr

    @pstderr.setter
    def pstderr(self, x):
        self._parameter_bucket.pstderr = x

    def set_cap(self, cap=25):
        """
        Set limiting values for one or more parameters.

        Parameters
        ----------
        cap : numeric, default 25.0
            Set a global limit on parameters.  The maximum has a ceiling
            at this value, and the minimum a floor at the negative of this, unless
            the existing bounds are entirely outside this range.
        """
        self.unmangle()
        return self._parameter_bucket.set_cap(cap=cap)

    def plock(self, values=None, **kwargs):
        self.unmangle()
        self._parameter_bucket.lock(values, **kwargs)

    def get_param_loc(self, name):
        self.unmangle()
        return self._parameter_bucket.get_param_loc(name)

    def get_value(self, name, *, default=None, kind="value"):
        if name is None and default is not None:
            return default
        if isinstance(name, dict):
            return {k: self.get_value(v) for k, v in name.items()}
        try:
            from .linear import LinearComponent, ParameterRef
            from .linear_math import ParameterOp

            if isinstance(name, (ParameterRef | ParameterOp)):
                return name.value(self)
            if isinstance(name, (LinearComponent | LinearFunction)):
                return name.as_pmath().value(self)
            if kind == "value":
                return self.pvals[self.get_param_loc(name)]
            return self.parameters[kind].values[self.get_param_loc(name)]
        except KeyError:
            if default is not None:
                return default
            else:
                raise

    @property
    def pf(self):
        self.unmangle()
        cols = ["value", "best", "initvalue", "minimum", "maximum", "nullvalue"]
        cols = [i for i in cols if i in self._parameter_bucket._params]
        return self._parameter_bucket._params[cols].to_dataframe()

    def set_values(self, values=None, **kwargs):
        """
        Set the parameter values for one or more parameters.

        Parameters
        ----------
        values : {'null', 'init', 'best', array-like, dict, scalar}, optional
            New values to set for the parameters.
            If 'null' or 'init', the current values are set
            equal to the null or initial values given in
            the 'nullvalue' or 'initvalue' column of the
            parameter frame, respectively.
            If 'best', the current values are set equal to
            the values given in the 'best' column of the
            parameter frame, if that columns exists,
            otherwise a ValueError exception is raised.
            If given as array-like, the array must be a
            vector with length equal to the length of the
            parameter frame, and the given vector will replace
            the current values.  If given as a dictionary,
            the dictionary is used to update `kwargs` before
            they are processed.
        kwargs : dict
            Any keyword arguments (or if `values` is a
            dictionary) are used to update the included named
            parameters only.  A warning will be given if any key of
            the dictionary is not found among the existing named
            parameters in the parameter frame, and the value
            associated with that key is ignored.  Any parameters
            not named by key in this dictionary are not changed.

        Notes
        -----
        Setting parameters both in the `values` argument and
        through keyword assignment is not explicitly disallowed,
        although it is not recommended.

        """
        warnings.warn(
            "Model.set_values(x) is deprecated, use Model.pvals = x",
            DeprecationWarning,
            stacklevel=2,
        )
        if isinstance(values, dict):
            kwargs.update(values)
        elif values is None:
            self.pvals = kwargs
        else:
            self.pvals = values

    def lock_value(self, name=None, value=None, **kwargs):
        """
        Set a fixed value for a model parameter.

        Parameters with a fixed value (i.e., with "holdfast" set to 1)
        will not be changed during estimation by the likelihood
        maximization algorithm.

        Parameters
        ----------
        name : str
            The name of the parameter to set to a fixed value.
        value : float
            The numerical value to set for the parameter.
        **kwargs
            Alternatively, use (name=value, ...) form for multiple
            parameters.
        """
        from .linear import ParameterRef

        if name is None or value is None:
            if name is not None or value is not None or len(kwargs) == 0:
                raise ValueError("give name and value or keyword arguments")
            for k, v in kwargs.items():
                self.lock_value(k, v)
            return

        if isinstance(name, ParameterRef):
            name = str(name)
        if value == "null":
            value = self.pf.loc[name, "nullvalue"]
        self.pvals = {name: value}
        self.pholdfast = {name: 0}
        self.pnullvals = {name: value}
        self.pminimum = {name: value}
        self.pmaximum = {name: value}
        self.pholdfast = {name: 1}

    def pretty_table(self):
        self.unmangle()
        return self._parameter_bucket.pretty_table()

    def __p_rename(self, x):
        return str(self.rename_parameters.get(str(x), str(x)))

    def __repr__(self):
        s = "<larch."
        s += self.__class__.__name__
        if self.is_mnl():
            s += " (MNL)"
        else:
            s += " (GEV)"
        if self.title != "Untitled":
            s += f' "{self.title}"'
        s += ">"
        return s

    def initialize_graph(
        self, alternative_codes=None, alternative_names=None, root_id=0
    ):
        """
        Write a nesting tree graph for a MNL model.

        Parameters
        ----------
        alternative_codes : array-like, optional
            Explicitly give alternative codes. Ignored if `dataframes` is given
            or if the model has dataframes or a dataservice already set.
        alternative_names : array-like, optional
            Explicitly give alternative names. Ignored if `dataframes` is given
            or if the model has dataframes or a dataservice already set.
        root_id : int, default 0
            The id code of the root node.

        Raises
        ------
        ValueError
            The model is unable to infer the alternative codes to use.  This can
            be avoided by giving alternative codes explicitly or having previously
            set dataframes or a dataservice that will give the alternative codes.
        """
        if self.datatree is not None:

            def get_coords_array(*names):
                for name in names:
                    if name in self.datatree.root_dataset.coords:
                        return self.datatree.root_dataset.coords[name].values

            if alternative_codes is None:
                alternative_codes = get_coords_array(
                    self.datatree.ALTID,
                    "_altid_",
                    "altid",
                    "alt_id",
                    "alt_ids",
                    "alternative_id",
                    "alternative_ids",
                )
            if alternative_names is None:
                alternative_names = get_coords_array(
                    "altname",
                    "altnames",
                    "alt_name",
                    "alt_names",
                    "alternative_name",
                    "alternative_names",
                )

        if alternative_codes is None:
            return

        from .tree import NestingTree

        g = NestingTree(root_id=root_id)
        if alternative_names is None:
            for a in alternative_codes:
                g.add_node(a)
        else:
            for a, name in zip(alternative_codes, alternative_names):
                g.add_node(a, name=name)
        self.graph = g

    @property
    def graph(self):
        if self._graph is None:
            try:
                self.initialize_graph()
            except ValueError as err:
                import warnings

                warnings.warn(
                    "cannot initialize graph, must define alternatives somehow",
                    stacklevel=2,
                )
                raise RuntimeError(
                    "cannot initialize graph, must define alternatives somehow"
                ) from err
        return self._graph

    @graph.setter
    def graph(self, x):
        self._graph = x

    def utility_functions(self, subset=None, resolve_parameters=False):
        """
        Generate an XHTML output of the utility function(s).

        Parameters
        ----------
        subset : Collection, optional
            A collection of alternative codes to include. This only has effect if
            there are separate utility_co functions set by alternative. It is
            recommended to use this parameter if there are a very large number of
            alternatives, and the utility functions of most (or all) of them
            can be effectively communicated by showing only a few.
        resolve_parameters : bool, default False
            Whether to resolve the parameters to the current (estimated) value
            in the output.

        Returns
        -------
        xmle.Elem
        """
        self.unmangle()
        from xmle import Elem

        x = Elem("div")
        t = x.elem("table", style="margin-top:1px;", attrib={"class": "floatinghead"})
        if len(self.utility_co):
            # t.elem('caption', text=f"Utility Functions",
            # 	   style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;"
            # 			 "font-style:normal;font-size:100%;padding:0px;color:black;")

            # iterate over all alternatives if a dataframes is attached and lists the alternatives
            try:
                if self.dataservice is not None:
                    alts = self.dataservice.alternative_codes()
                elif self.dataframes is not None:
                    alts = self.dataframes.alternative_codes()
                else:
                    alts = self.utility_co.keys()
            except Exception:
                alts = self.utility_co.keys()
            t_head = t.elem("thead")
            tr = t_head.elem("tr")
            tr.elem("th", text="alt")
            tr.elem("th", text="formula", attrib={"style": "text-align:left;"})
            t_body = t.elem("tbody")
            for j in alts:
                if subset is None or j in subset:
                    tr = t_body.elem("tr")
                    tr.elem("td", text=str(j))
                    utilitycell = tr.elem("td", attrib={"style": "text-align:left;"})
                    utilitycell.elem("div")
                    anything = False
                    if len(self.utility_ca):
                        utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
                        utilitycell << list(
                            self.utility_ca.__xml__(
                                linebreaks=True,
                                resolve_parameters=self,
                                value_in_tooltips=not resolve_parameters,
                            )
                        )
                        anything = True
                    if j in self.utility_co:
                        v = self.utility_co[j]
                        if len(v):
                            if anything:
                                utilitycell << Elem("br")
                            utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
                            utilitycell << list(
                                v.__xml__(
                                    linebreaks=True,
                                    resolve_parameters=self,
                                    value_in_tooltips=not resolve_parameters,
                                )
                            )
                            anything = True
                    if len(self.quantity_ca):
                        if anything:
                            utilitycell << Elem("br")
                        if self.quantity_scale:
                            utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
                            from .linear import ParameterRef

                            utilitycell << list(
                                ParameterRef(self.quantity_scale).__xml__(
                                    resolve_parameters=self,
                                    value_in_tooltips=not resolve_parameters,
                                )
                            )
                            utilitycell[-1].tail = (
                                utilitycell[-1].tail or ""
                            ) + " * log("
                        else:
                            utilitycell[-1].tail = (
                                utilitycell[-1].tail or ""
                            ) + " + log("
                        content = self.quantity_ca.__xml__(
                            linebreaks=True,
                            lineprefix="  ",
                            exponentiate_parameters=True,
                            resolve_parameters=self,
                            value_in_tooltips=not resolve_parameters,
                        )
                        utilitycell << list(content)
                        utilitycell.elem("br", tail=")")
        else:
            # there is no differentiation by alternatives, just give one formula
            # t.elem('caption', text=f"Utility Function",
            # 	   style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;"
            # 			 "font-style:normal;font-size:100%;padding:0px;color:black;")
            tr = t.elem("tr")
            utilitycell = tr.elem("td", attrib={"style": "text-align:left;"})
            utilitycell.elem("div")
            anything = False
            if len(self.utility_ca):
                utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
                utilitycell << list(
                    self.utility_ca.__xml__(
                        linebreaks=True,
                        resolve_parameters=self,
                        value_in_tooltips=not resolve_parameters,
                    )
                )
                anything = True
            if len(self.quantity_ca):
                if anything:
                    utilitycell << Elem("br")
                if self.quantity_scale:
                    utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + "
                    from .linear import ParameterRef

                    utilitycell << list(
                        ParameterRef(self.quantity_scale).__xml__(
                            resolve_parameters=self,
                            value_in_tooltips=not resolve_parameters,
                        )
                    )
                    utilitycell[-1].tail = (utilitycell[-1].tail or "") + " * log("
                else:
                    utilitycell[-1].tail = (utilitycell[-1].tail or "") + " + log("
                content = self.quantity_ca.__xml__(
                    linebreaks=True,
                    lineprefix="  ",
                    exponentiate_parameters=True,
                    resolve_parameters=self,
                    value_in_tooltips=not resolve_parameters,
                )
                utilitycell << list(content)
                utilitycell.elem("br", tail=")")
        return x

    def _utility_functions_as_frame(self, subset=None, resolve_parameters=False):
        """
        Generate a tabular output of the utility function(s).

        Parameters
        ----------
        subset : Collection, optional
            A collection of alternative codes to include. This only has effect if
            there are separate utility_co functions set by alternative. It is
            recommended to use this parameter if there are a very large number of
            alternatives, and the utility functions of most (or all) of them
            can be effectively communicated by showing only a few.
        resolve_parameters : bool, default False
            Whether to resolve the parameters to the current (estimated) value
            in the output.  Not implemented.

        Returns
        -------
        xmle.Elem
        """
        self.unmangle()

        tf = pd.DataFrame(
            index=pd.MultiIndex.from_tuples([], names=["Alt", "Line"]),
            columns=["Formula"],
        )

        if len(self.utility_co):
            # iterate over all alternatives if a dataframes is attached and lists the alternatives
            try:
                if self.dataservice is not None:
                    alts = self.dataservice.alternative_codes()
                elif self.dataframes is not None:
                    alts = self.dataframes.alternative_codes()
                else:
                    alts = self.utility_co.keys()
            except Exception:
                alts = self.utility_co.keys()

            for a in alts:
                if not (subset is None or a in subset):
                    continue
                line = 1
                op = " "
                for i in self.utility_ca:
                    tf.loc[(a, line), :] = f"{op} {str(i)}"
                    op = "+"
                    line += 1
                if a in self.utility_co:
                    for i in self.utility_co[a]:
                        tf.loc[(a, line), :] = f"{op} {str(i)}"
                        op = "+"
                        line += 1
                if len(self.quantity_ca):
                    if self.quantity_scale:
                        from .linear import ParameterRef

                        q = ParameterRef(self.quantity_scale)
                        tf.loc[(a, line), :] = f"{op} {str(q)} * log("
                    else:
                        tf.loc[(a, line), :] = f"{op} log("
                    op = " "
                    line += 1
                    for i in self.quantity_ca:
                        tf.loc[(a, line), :] = f"    {op} {i._str_exponentiate()}"
                        op = "+"
                        line += 1
                    tf.loc[(a, line), :] = ")"
                    op = "+"
                    line += 1

        else:
            # there is no differentiation by alternatives, just give one formula
            a = "*"
            line = 1
            op = " "
            for i in self.utility_ca:
                tf.loc[(a, line), :] = f"{op} {str(i)}"
                op = "+"
                line += 1
            if len(self.quantity_ca):
                if self.quantity_scale:
                    from .linear import ParameterRef

                    q = ParameterRef(self.quantity_scale)
                    tf.loc[(a, line), :] = f"{op} {str(q)} * log("
                else:
                    tf.loc[(a, line), :] = f"{op} log("
                op = " "
                line += 1
                for i in self.quantity_ca:
                    tf.loc[(a, line), :] = f"    {op} {i._str_exponentiate()}"
                    op = "+"
                    line += 1
                tf.loc[(a, line), :] = ")"
                op = "+"
                line += 1

        return tf

    def required_data(self):
        """
        Report what data is required in DataFrames for this model to be used.

        Returns
        -------
        dictx
        """
        try:
            from ..util import dictx

            req_data = dictx()

            if self.utility_ca is not None and len(self.utility_ca):
                if "ca" not in req_data:
                    req_data.ca = set()
                for i in self.utility_ca:
                    req_data.ca.add(str(i.data))

            if self.quantity_ca is not None and len(self.quantity_ca):
                if "ca" not in req_data:
                    req_data.ca = set()
                for i in self.quantity_ca:
                    req_data.ca.add(str(i.data))

            if self.utility_co is not None and len(self.utility_co):
                if "co" not in req_data:
                    req_data.co = set()
                for _alt, func in self.utility_co.items():
                    for i in func:
                        if str(i.data) != "1":
                            req_data.co.add(str(i.data))

            if "ca" in req_data:
                req_data.ca = list(sorted(req_data.ca))
            if "co" in req_data:
                req_data.co = list(sorted(req_data.co))

            if self.choice_ca_var:
                req_data.choice_ca = self.choice_ca_var
            elif self.choice_co_vars:
                req_data.choice_co = self.choice_co_vars
            elif self.choice_co_code:
                req_data.choice_co_code = self.choice_co_code
            elif self.choice_any:
                req_data.choice_any = True

            if self.weight_co_var:
                req_data.weight_co = self.weight_co_var

            if self.availability_ca_var:
                req_data.avail_ca = self.availability_ca_var
            elif self.availability_co_vars:
                req_data.avail_co = self.availability_co_vars
            elif self.availability_any:
                req_data.avail_any = True

            return req_data
        except Exception:
            logger.exception("error in required_data")

    def __contains__(self, item):
        return item in self.pf.index  # or (item in self.rename_parameters)

    @property
    def is_mangled(self):
        return self._mangled

    def mangle(self, data=True, structure=True):
        if data and not (self._mangled & 0x1):
            logger.debug(f"mangling data for {self.title}")
            self._mangled |= MANGLE_DATA
        if structure and not (self._mangled & MANGLE_STRUCTURE):
            logger.debug(f"mangling structure for {self.title}")
            self._mangled |= MANGLE_STRUCTURE

    def unmangle(self, force=False, structure_only=False):
        if not self._mangled and not force:
            return
        marker = f"_currently_unmangling_{__file__}"
        logger.debug(f"{marker=}")
        if getattr(self, marker, False):
            return
        try:
            setattr(self, marker, True)
            if force:
                self.mangle()
            logger.debug(f"{self._mangled=}")
            if (self._mangled & MANGLE_STRUCTURE) or force:
                self._scan_all_ensure_names()
                if structure_only:
                    self._mangled &= ~MANGLE_STRUCTURE
                else:
                    self._mangled = 0
        finally:
            delattr(self, marker)

    def _scan_all_ensure_names(self):
        self._scan_utility_ensure_names()
        self._scan_quantity_ensure_names()
        self._scan_logsums_ensure_names()
        self._scan_mixtures_ensure_names()

    def _scan_utility_ensure_names(self):
        """
        Scan the utility functions and ensure all named parameters appear in the parameter frame.

        Any named parameters that do not appear in the parameter frame are added.
        """
        nameset = set()
        u_co_dataset = set()
        for altcode, linear_function in self.utility_co.items():
            for component in linear_function:
                nameset.add(self.__p_rename(component.param))
                try:
                    u_co_dataset.add(str(component.data))
                except Exception:
                    warnings.warn(f"bad data in altcode {altcode}", stacklevel=2)
                    raise
        linear_function_ca = self.utility_ca
        for component in linear_function_ca:
            nameset.add(self.__p_rename(component.param))
        self._ensure_names(nameset)

    def _scan_quantity_ensure_names(self):
        if self.quantity_ca is not None:
            nameset = set()
            for component in self.quantity_ca:
                nameset.add(self.__p_rename(component.param))
            self._ensure_names(nameset)

    def _scan_logsums_ensure_names(self):
        nameset = set()
        if self._graph is not None:
            for nodecode in self._graph.topological_sorted_no_elementals:
                if nodecode != self._graph._root_id:
                    param_name = str(self._graph.nodes[nodecode]["parameter"])
                    nameset.add(self.__p_rename(param_name))
        if self.quantity_ca is not None and len(self.quantity_ca) > 0:
            if self.quantity_scale is not None:
                nameset.add(self.__p_rename(self.quantity_scale))
        if self.logsum_parameter is not None:
            nameset.add(self.__p_rename(self.logsum_parameter))
        self._ensure_names(
            nameset, value=1, nullvalue=1, initvalue=1, minimum=0.001, maximum=1
        )

    def _scan_mixtures_ensure_names(self):
        for i in self.mixtures:
            for name, default_value in i.param_names().items():
                self._ensure_names([name], value=default_value, initvalue=default_value)

    def _ensure_names(self, names, **kwargs):
        if self._parameter_bucket is not None:
            self._parameter_bucket.add_parameters(names, fill_values=kwargs)

    quantity_scale = SingleParameter()
    logsum_parameter = SingleParameter()

    def clear_cache(self):
        """Remove all cached log likelihood values and estimation results."""
        self._cached_loglike_best = None
        self._cached_loglike_null = None
        self._most_recent_estimation_result = None

    def _check_if_best(self, computed_ll, pvalues=None):
        if self._cached_loglike_best is None or computed_ll > self._cached_loglike_best:
            self._cached_loglike_best = computed_ll
            if pvalues is None:
                self._parameter_bucket._params = self._parameter_bucket._params.assign(
                    best=self._parameter_bucket._params["value"]
                )
            else:
                self._parameter_bucket._params = self._parameter_bucket._params.assign(
                    best=xr.DataArray(pvalues, dims=self._parameter_bucket.index_name)
                )

    @property
    def choice_ca_var(self):
        """Str : An |idca| variable giving the choices as indicator values."""
        return self._choice_ca_var

    @choice_ca_var.setter
    def choice_ca_var(self, x):
        if x is not None:
            x = str(x)
        if self._choice_ca_var != x:
            self.mangle()
        self._choice_ca_var = x
        if x is not None:
            self._choice_co_vars = None
            self._choice_co_code = None
            self._choice_any = False

    @property
    def choice_co_vars(self):
        """Dict[int,str] : A mapping giving |idco| expressions that evaluate to indicator values.

        Each key represents an alternative code number, and the associated expression
        gives the name of an |idco| variable or some function of |idco| variables that
        indicates whether that alternative was chosen.
        """
        if self._choice_co_vars:
            return self._choice_co_vars
        else:
            return None

    @choice_co_vars.setter
    def choice_co_vars(self, x):
        if isinstance(x, dict):
            if self._choice_co_vars != x:
                self.mangle()
            self._choice_co_vars = x
            self._choice_ca_var = None
            self._choice_co_code = None
            self._choice_any = False
        elif x is None:
            if self._choice_co_vars != x:
                self.mangle()
            self._choice_co_vars = x
        else:
            raise TypeError("choice_co_vars must be a dictionary")

    @choice_co_vars.deleter
    def choice_co_vars(self):
        self._choice_co_vars = None

    @property
    def choice_co_code(self):
        """Str : An |idco| variable giving the choices as alternative id's."""
        if self._choice_co_code:
            return self._choice_co_code
        else:
            return None

    @choice_co_code.setter
    def choice_co_code(self, x):
        if isinstance(x, str):
            if self._choice_co_code != x:
                self.mangle()
            self._choice_co_code = x
            self._choice_co_vars = None
            self._choice_ca_var = None
            self._choice_any = False
        elif x is None:
            if self._choice_co_code != x:
                self.mangle()
            self._choice_co_code = x
        else:
            raise TypeError("choice_co_vars must be a str")

    @choice_co_code.deleter
    def choice_co_code(self):
        if self._choice_co_code is not None:
            self.mangle()
        self._choice_co_code = None

    @property
    def choice_any(self):
        if self._choice_any:
            return True
        else:
            return False

    @choice_any.setter
    def choice_any(self, x):
        if x:
            self._choice_any = True
            self._choice_co_code = None
            self._choice_co_vars = None
            self._choice_ca_var = None
        else:
            self._choice_any = False

    @choice_any.deleter
    def choice_any(self):
        self._choice_any = False

    @property
    def weight_co_var(self):
        return self._weight_co_var

    @weight_co_var.setter
    def weight_co_var(self, x):
        self._weight_co_var = x

    @property
    def availability_ca_var(self):
        """Str : An |idca| variable or expression indicating if alternatives are available."""
        return self._availability_ca_var

    @availability_ca_var.setter
    def availability_ca_var(self, x):
        if x is not None:
            x = str(x)
        if self._availability_ca_var != x:
            self.mangle()
            self._availability_ca_var = x
            self._availability_co_vars = None
            self._availability_any = False

    @property
    def availability_co_vars(self):
        """Dict[int,str] : A mapping giving |idco| expressions that evaluate to availability indicators.

        Each key represents an alternative code number, and the associated expression
        gives the name of an |idco| variable or some function of |idco| variables that
        indicates whether that alternative is available.
        """
        if self._availability_co_vars:
            return self._availability_co_vars
        else:
            return None

    @availability_co_vars.setter
    def availability_co_vars(self, x):
        from collections.abc import Mapping

        if x is None:
            if self._availability_co_vars:
                self.mangle()
                self._availability_co_vars = {}
        else:
            if not isinstance(x, Mapping):
                raise TypeError(f"availability_co_vars must be dict not {type(x)}")
            if self._availability_co_vars != x:
                self.mangle()
            self._availability_co_vars = x
            self._availability_ca_var = None
            self._availability_any = False

    @property
    def availability_any(self):
        """Bool : A flag indicating whether availability should be inferred from the data.

        This only applies to DataFrames-based models, as the Dataset interface does
        not include a mechanism for the data to self-describe an availability feature.
        """
        return self._availability_any

    @availability_any.setter
    def availability_any(self, x):
        x = bool(x)
        if x != self._availability_any:
            self._availability_any = x
            self._availability_co_vars = None
            self._availability_ca_var = None

    def parameter_summary(self):
        """
        Create a tabular summary of parameter values.

        This will generate a small table of parameters statistics,
        containing:

        *	Parameter Name (and Category, if applicable)
        *	Estimated Value
        *	Standard Error of the Estimate (if known)
        *	t Statistic (if known)
        *	Null Value
        *	Binding Constraints (if applicable)

        Returns
        -------
        pandas.DataFrame

        """
        NBSP = " "  # non=breaking space

        try:
            pbucket = self._parameter_bucket
        except AttributeError:
            pbucket = self

        tabledata = {}
        tabledata["Value"] = pbucket.pvals
        if "std_err" in pbucket._params:
            se = pbucket.pstderr
            tabledata["Std Err"] = se
            tstat = (pbucket.pvals - pbucket.pnullvals) / np.where(se, se, 1.0)
            tabledata["t Stat"] = np.where(se, tstat, np.nan)
        if "robust_std_err" in pbucket._params:
            se_ = pbucket._params["robust_std_err"]
            tabledata["Robust Std Err"] = se_
            tstat_ = (pbucket.pvals - pbucket.pnullvals) / np.where(se_, se_, 1.0)
            tabledata["Robust t Stat"] = np.where(se_, tstat_, np.nan)
        tabledata["Null Value"] = pbucket.pnullvals

        if "constrained" in pbucket.parameters:
            tabledata["Constrained"] = pbucket.parameters["constrained"]

        result = pd.DataFrame(tabledata, index=pbucket.pnames).rename_axis(
            index="Parameter"
        )

        # pf = self.pf
        # columns = [i for i in
        #           ['value', 'std_err', 't_stat', 'likelihood_ratio', 'nullvalue', 'constrained']
        #           if i in pf.columns]
        # result = pf[columns].rename(
        #     columns={
        #         'value': 'Value',
        #         'std_err': 'Std Err',
        #         't_stat': 't Stat',
        #         'likelihood_ratio': 'Like Ratio',
        #         'nullvalue': 'Null Value',
        #         'constrained': 'Constrained'
        #     }
        # )
        monospace_cols = []

        def fixie(x):
            if np.isfinite(x):
                if x > 1000:
                    return NBSP + "BIG"
                elif x < -1000:
                    return "-BIG"
                else:
                    return f"{x:0< 4.2f}".replace(" ", NBSP)
            else:
                return NBSP + "NA"

        if "t Stat" in result.columns:
            result.insert(result.columns.get_loc("t Stat") + 1, "Signif", "")
            result.loc[np.absolute(result["t Stat"]) > 1.9600, "Signif"] = "*"
            result.loc[np.absolute(result["t Stat"]) > 2.5758, "Signif"] = "**"
            result.loc[np.absolute(result["t Stat"]) > 3.2905, "Signif"] = "***"
            result["t Stat"] = result["t Stat"].apply(fixie)
            result.loc[result["t Stat"] == NBSP + "NA", "Signif"] = ""
            monospace_cols.append("t Stat")
            monospace_cols.append("Signif")
        if "Robust t Stat" in result.columns:
            result.insert(
                result.columns.get_loc("Robust t Stat") + 1, "Robust Signif", ""
            )
            result.loc[
                np.absolute(result["Robust t Stat"]) > 1.9600, "Robust Signif"
            ] = "*"
            result.loc[
                np.absolute(result["Robust t Stat"]) > 2.5758, "Robust Signif"
            ] = "**"
            result.loc[
                np.absolute(result["Robust t Stat"]) > 3.2905, "Robust Signif"
            ] = "***"
            result["Robust t Stat"] = result["Robust t Stat"].apply(fixie)
            result.loc[result["Robust t Stat"] == NBSP + "NA", "Robust Signif"] = ""
            monospace_cols.append("Robust t Stat")
            monospace_cols.append("Robust Signif")
        if "Like Ratio" in result.columns:
            if "Signif" not in result.columns:
                result.insert(result.columns.get_loc("Like Ratio") + 1, "Signif", "")
            if "t Stat" in result.columns:
                non_finite_t = ~np.isfinite(result["t Stat"])
            else:
                non_finite_t = True
            result.loc[
                np.absolute((np.isfinite(result["Like Ratio"])) & non_finite_t),
                "Signif",
            ] = "[]"
            result.loc[
                np.absolute(((result["Like Ratio"]) > 1.9207) & non_finite_t), "Signif"
            ] = "[*]"
            result.loc[
                np.absolute(((result["Like Ratio"]) > 3.3174) & non_finite_t), "Signif"
            ] = "[**]"
            result.loc[
                np.absolute(((result["Like Ratio"]) > 5.4138) & non_finite_t), "Signif"
            ] = "[***]"
            result["Like Ratio"] = result["Like Ratio"].apply(fixie)
            monospace_cols.append("Like Ratio")
            if "Signif" not in monospace_cols:
                monospace_cols.append("Signif")
        for z in ["Std Err", "Robust Std Err"]:
            if z in result.columns:
                _fmt_s = (
                    lambda x: f"{x: #.3g}".replace(" ", NBSP)
                    if np.isfinite(x)
                    else NBSP + "NA"
                )
                result[z] = result[z].apply(_fmt_s)
                monospace_cols.append(z)
        if "Value" in result.columns:
            result["Value"] = result["Value"].apply(
                lambda x: f"{x: #.3g}".replace(" ", NBSP)
            )
            monospace_cols.append("Value")
        if "Constrained" in result.columns:
            result["Constrained"] = result["Constrained"].str.replace("\n", "<br/>")
        if "Null Value" in result.columns:
            monospace_cols.append("Null Value")
        # if result.index.nlevels > 1:
        #     pnames = result.index.get_level_values(-1)
        # else:
        #     pnames = result.index
        styles = [
            dict(
                selector="th",
                props=[
                    ("vertical-align", "top"),
                    ("text-align", "left"),
                ],
            ),
            dict(
                selector="td",
                props=[
                    ("vertical-align", "top"),
                    ("text-align", "left"),
                ],
            ),
        ]

        if self.ordering:
            paramset = set(result.index)
            out = []
            import re

            for category in self.ordering:
                category_name = category[0]
                category_params = []
                for category_pattern in category[1:]:
                    category_params.extend(
                        sorted(
                            i
                            for i in paramset
                            if re.match(category_pattern, i) is not None
                        )
                    )
                    paramset -= set(category_params)
                out.append([category_name, category_params])
            if len(paramset):
                out.append(["Other", sorted(paramset)])

            tuples = []
            for c, pp in out:
                for p in pp:
                    tuples.append((c, p))

            ix = pd.MultiIndex.from_tuples(tuples, names=["Category", "Parameter"])

            result = result.reindex(ix.get_level_values(1))
            result.index = ix

        return (
            result.style.set_table_styles(styles)
            .format({"Null Value": "{: .2f}"})
            .map(lambda x: "font-family:monospace", subset=monospace_cols)
        )

    def estimation_statistics(self, compute_loglike_null=True):
        """
        Create an XHTML summary of estimation statistics.

        This will generate a small table of estimation statistics,
        containing:

        *	Log Likelihood at Convergence
        *	Log Likelihood at Null Parameters (if known)
        *	Log Likelihood with No Model (if known)
        *	Log Likelihood at Constants Only (if known)

        Additionally, for each included reference value (i.e.
        everything except log likelihood at convergence) the
        rho squared with respect to that value is also given.

        Each statistic is reported in aggregate, as well as
        per case.

        Parameters
        ----------
        compute_loglike_null : bool, default True
            If the log likelihood at null values has not already
            been computed (i.e., if it is not cached) then compute
            it, cache its value, and include it in the output.

        Returns
        -------
        xmle.Elem

        """
        from xmle import Elem

        div = Elem("div")
        table = div.put("table")

        thead = table.put("thead")
        tr = thead.put("tr")
        tr.put("th", text="Statistic")
        tr.put("th", text="Aggregate")
        tr.put("th", text="Per Case")

        tbody = table.put("tbody")

        try:
            ncases = self.n_cases
        except (MissingDataError, AttributeError):
            ncases = None

        tr = tbody.put("tr")
        tr.put("td", text="Number of Cases")
        if ncases:
            tr.put("td", text=str(ncases), colspan="2")
        else:
            tr.put("td", text="not available", colspan="2")

        mostrecent = self._most_recent_estimation_result
        if mostrecent is not None:
            tr = tbody.put("tr")
            tr.put("td", text="Log Likelihood at Convergence")
            tr.put("td", text=f"{mostrecent.loglike:.2f}")
            if ncases:
                tr.put("td", text=f"{mostrecent.loglike / ncases:.2f}")
            else:
                tr.put("td", text="na")

        ll_z = self._cached_loglike_null
        if ll_z == 0 or ll_z is None:
            if compute_loglike_null:
                try:
                    ll_z = self.loglike_null()
                except (MissingDataError, AttributeError):
                    ll_z = 0
            else:
                ll_z = 0
        if ll_z:
            tr = tbody.put("tr")
            tr.put("td", text="Log Likelihood at Null Parameters")
            tr.put("td", text=f"{ll_z:.2f}")
            if ncases:
                tr.put("td", text=f"{ll_z / ncases:.2f}")
            else:
                tr.put("td", text="na")
            if mostrecent is not None:
                tr = tbody.put("tr")
                tr.put("td", text="Rho Squared w.r.t. Null Parameters")
                rsz = 1.0 - (mostrecent.loglike / ll_z)
                tr.put("td", text=f"{rsz:.3f}", colspan="2")

        try:
            ll_nil = self._cached_loglike_nil
        except (MissingDataError, AttributeError):
            ll_nil = 0
        if ll_nil:
            tr = tbody.put("tr")
            tr.put("td", text="Log Likelihood with No Model")
            tr.put("td", text=f"{ll_nil:.2f}")
            if ncases:
                tr.put("td", text=f"{ll_nil / ncases:.2f}")
            else:
                tr.put("td", text="na")
            if mostrecent is not None:
                tr = tbody.put("tr")
                tr.put("td", text="Rho Squared w.r.t. No Model")
                rsz = 1.0 - (mostrecent.loglike / ll_nil)
                tr.put("td", text=f"{rsz:.3f}", colspan="2")

        try:
            ll_c = self._cached_loglike_constants_only
        except (MissingDataError, AttributeError):
            ll_c = 0
        if ll_c:
            tr = tbody.put("tr")
            tr.put("td", text="Log Likelihood at Constants Only")
            tr.put("td", text=f"{ll_c:.2f}")
            if ncases:
                tr.put("td", text=f"{ll_c / ncases:.2f}")
            else:
                tr.put("td", text="na")
            if mostrecent is not None:
                tr = tbody.put("tr")
                tr.put("td", text="Rho Squared w.r.t. Constants Only")
                rsc = 1.0 - (mostrecent.loglike / ll_c)
                tr.put("td", text=f"{rsc:.3f}", colspan="2")

        return div



import numpy as np

from ..exceptions import BHHHSimpleStepFailure


def propose_direction(bhhh, dloglike, freedoms):
    direction = np.zeros_like(dloglike)
    try:
        ibhhh = np.linalg.inv(bhhh[freedoms, :][:, freedoms])
    except np.linalg.LinAlgError:
        ibhhh = np.linalg.pinv(bhhh[freedoms, :][:, freedoms])
    direction1 = np.dot(ibhhh, dloglike[freedoms])
    # direction1 = np.linalg.lstsq(
    #     bhhh[freedoms, :][:, freedoms],
    #     dloglike[freedoms],
    #     rcond=None,
    # )[0]
    direction[freedoms] = direction1
    return direction


def fit_bhhh(
    model,
    steplen=1.0,
    momentum=5,
    logger=None,
    ctol=1e-4,
    maxiter=100,
    soft_maxiter=None,
    callback=None,
    minimum_steplen=0.0001,
    maximum_steplen=1.0,
    leave_out=-1,
    keep_only=-1,
    subsample=-1,
    initial_constraint_intensity=None,
    step_constraint_intensity=1.5,
    max_constraint_intensity=1e6,
    initial_constraint_sharpness=None,
    step_constraint_sharpness=1.5,
    max_constraint_sharpness=1e6,
    jumpstart=0,
    jumpstart_split=5,
):
    """
    Make a series of steps using the BHHH algorithm.

    Parameters
    ----------
    steplen: float
    logger: logging.Logger

    Returns
    -------
    loglike, convergence_tolerance, n_iters, steps
    """
    if logger is None:

        class NoLogger:
            debug = lambda *x: None
            info = lambda *x: None

        logger = NoLogger()

    iter = 0
    steps = []

    if initial_constraint_intensity is not None:
        model.constraint_intensity = initial_constraint_intensity
    if initial_constraint_sharpness is not None:
        model.constraint_sharpness = initial_constraint_sharpness

    if jumpstart:
        current_result = model.loglike2_bhhh(
            leave_out=leave_out,
            keep_only=keep_only,
            subsample=subsample,
        )
        current_ll = current_result.ll
        logger.debug(f"before jumpstart loglike {current_ll}")

        model.jumpstart_bhhh(
            jumpstart=jumpstart,
            jumpstart_split=jumpstart_split,
            logger=logger,
        )
        iter += jumpstart

    current_pvals = model.pvals.copy()
    current_result = model.loglike2_bhhh(
        leave_out=leave_out,
        keep_only=keep_only,
        subsample=subsample,
    )
    current_ll = current_result.ll
    current_dll = current_result.dll
    current_bhhh = current_result.bhhh

    logger.debug(f"initial loglike {current_ll}")

    def find_direction(current_dll, current_bhhh):
        freedoms = model.pholdfast == 0
        direction = propose_direction(current_bhhh, np.asarray(current_dll), freedoms)
        tolerance = np.dot(direction, current_dll)
        return direction, tolerance

    direction, tolerance = find_direction(current_dll, current_bhhh)

    message = "Optimization terminated for undetermined reason."

    while True:
        # check various break conditions
        current_constraint_violation = model.constraint_violation(
            on_violation="return", intensity_check=True
        )
        if current_constraint_violation:
            logger.debug(
                f"current constraint violation: {current_constraint_violation}"
            )
        if abs(tolerance) <= ctol and not current_constraint_violation:
            message = "Optimization terminated successfully"
            break
        if iter >= maxiter:
            if current_constraint_violation:
                message = (
                    f"Optimization terminated after {iter} iterations "
                    f"with {current_constraint_violation}"
                )
            else:
                message = f"Optimization terminated after {iter} iterations"
            break
        if soft_maxiter is not None:
            if iter >= soft_maxiter and not current_constraint_violation:
                message = (
                    f"Optimization terminated after {iter} iterations "
                    f"with no constraint violations"
                )
                break

        # no break, make a step
        iter += 1
        if steps:
            steplen = min(
                2.0 * sum(steps[-momentum:]) / len(steps[-momentum:]), maximum_steplen
            )
        while True:
            model.pvals = current_pvals + direction * steplen
            proposed = model.loglike2_bhhh(
                leave_out=leave_out,
                keep_only=keep_only,
                subsample=subsample,
            )
            proposed_ll = proposed.ll
            if proposed_ll > current_ll:
                break
            logger.debug(
                f"failed simple step bhhh {steplen}, degraded {current_ll - proposed_ll}"
            )
            steplen *= 0.5
            if steplen < minimum_steplen:
                break
        if proposed_ll <= current_ll:
            logger.debug("no improvement found, reset to prior x")
            model.set_values(current_pvals)
            if not model.constraint_violation(on_violation="return"):
                logger.debug("-- no constraints are violated")
                # no violations, check that penalty is small
                if np.absolute(current_result.penalty * model.n_cases) > 0.01:
                    logger.debug(
                        f"-- penalty is not small ({current_result.penalty * model.n_cases}), "
                        f"increasing sharpness"
                    )
                    # penalty is not small, crank up sharpness
                    proposed.penalty = current_result.penalty
                    while (
                        np.absolute(proposed.penalty * model.n_cases) > 0.01
                        and model.constraint_sharpness < max_constraint_sharpness
                    ):
                        model.constraint_sharpness *= step_constraint_sharpness
                        model.constraint_sharpness = min(
                            model.constraint_sharpness, max_constraint_sharpness
                        )
                        logger.debug(f"-- sharpness to ({model.constraint_sharpness})")
                        proposed = model.loglike2_bhhh(
                            leave_out=leave_out,
                            keep_only=keep_only,
                            subsample=subsample,
                        )
                        proposed_ll = proposed.ll
                        logger.debug(
                            f"   penalty is ({proposed.penalty * model.n_cases})"
                        )
                    logger.debug(
                        f"-- proposed_ll {proposed_ll}, current_ll is {current_ll}, will continue"
                    )
                else:
                    logger.debug(
                        f"-- penalty is small ({current_result.penalty * model.n_cases}), "
                        f"checking for converge"
                    )
                    tolerance = model.constraint_converge_tolerance()
                    logger.debug(f"-- tolerance is {tolerance}")
                    if np.absolute(tolerance) < ctol:
                        message = "Constrained optimization terminated successfully."
                        logger.debug(message)
                        break
                    else:
                        logger.debug("-- not yet well converged")
            else:
                raise BHHHSimpleStepFailure(
                    f"simple step bhhh failed\ndirection = {str(direction)}"
                )
        else:
            logger.debug(
                f"simple step bhhh {steplen} gains {proposed_ll - current_ll} to {proposed_ll}"
            )
        steps.append(steplen)

        current_ll, current_dll, current_bhhh = proposed_ll, proposed.dll, proposed.bhhh
        current_pvals = model.pvals.copy()
        if callback is not None:
            callback(
                current_pvals,
                {
                    "penalty": proposed.penalty * model.n_cases,
                    "total_loglike": current_ll,
                    "actual_loglike": current_ll - proposed.penalty,
                },
            )

        if model.constraint_intensity and (
            model.constraint_intensity < max_constraint_intensity
            or model.constraint_sharpness < max_constraint_sharpness
        ):
            logger.debug("increasing constraint intensity and sharpness")
            model.constraint_intensity *= step_constraint_intensity
            model.constraint_sharpness *= step_constraint_sharpness
            model.constraint_intensity = min(
                model.constraint_intensity, max_constraint_intensity
            )
            model.constraint_sharpness = min(
                model.constraint_sharpness, max_constraint_sharpness
            )
            # recompute value with updated penalties
            current_result = model.loglike2_bhhh(
                leave_out=leave_out,
                keep_only=keep_only,
                subsample=subsample,
            )
            current_ll = current_result.ll
            current_dll = current_result.dll
            current_bhhh = current_result.bhhh
            logger.debug(
                f"  constrained convergence tolerance {model.constraint_converge_tolerance()}"
            )

        direction, tolerance = find_direction(current_dll, current_bhhh)
        logger.debug(f"  convergence tolerance {tolerance}")
        logger.debug(f"  constraint intensity {model.constraint_intensity}")
        logger.debug(f"  constraint sharpness {model.constraint_sharpness}")

    return current_ll, tolerance, iter, np.asarray(steps), message



import numpy


def define_masks(bitmask_sizes):
    """
    Define a list of bitmasks from a list of bitmask sizes.

    Parameters
    ----------
    bitmask_sizes : iterable

    Returns
    -------
    list
    """
    shifts = ((numpy.cumsum(bitmask_sizes[::-1]))[::-1]).astype(int)
    shifts[:-1] = shifts[1:]
    shifts[-1] = 0

    masks = [
        ((2 ** bitmask_sizes[b]) - 1) << shifts[b] for b in range(len(bitmask_sizes))
    ]

    return masks


def define_min_masks_from_sets(sets):
    """
    Define a list of bitmasks from a list of sets.

    Parameters
    ----------
    sets : iterable of iterable

    Returns
    -------
    list
    """
    bitmask_sizes = [int(numpy.ceil(numpy.log2(len(numpy.unique(s))))) for s in sets]
    return define_masks(bitmask_sizes)


def define_min_masks_from_df(df):
    """
    Define a list of bitmasks from a DataFrame.

    Parameters
    ----------
    df : pandas.DataFrame

    Returns
    -------
    list
    """
    bitmask_sizes = [
        int(numpy.ceil(numpy.log2(len(numpy.unique(df[s]))))) for s in df.columns
    ]
    return define_masks(bitmask_sizes)



from .display import HTML, display_html
from .xhtml import tooltipped_style

body_font = 'font-family: "Book-Antiqua", "Palatino", serif;'

signature_font = (
    "font-size:70%; font-weight:100; font-style:italic; "
    "font-family: Roboto, Helvetica, sans-serif;"
)

signature_name_font = (
    "font-weight:400; font-style:normal; "
    'font-family: "Roboto Slab", Roboto, Helvetica, sans-serif;'
)


def load_css(filename):
    import os

    css = None
    if filename is None or not isinstance(filename, str):
        return None
    if os.path.exists(filename):
        with open(filename) as f:
            css = f.read()
        return css
    f0 = f"{filename}.css"
    if os.path.exists(f0):
        with open(f0) as f:
            css = f.read()
        return css
    try:
        import platformdirs
    except ImportError:
        pass
    else:
        f1 = os.path.join(platformdirs.user_config_dir("Larch"), filename)
        if os.path.exists(f1):
            with open(f1) as f:
                css = f.read()
            return css
        f2 = f"{f1}.css"
        if os.path.exists(f2):
            with open(f2) as f:
                css = f.read()
            return css
    if "{" in filename and "}" in filename:
        return filename
    return css


_default_css_jupyter = (
    """

@import url('https://fonts.googleapis.com/css?family=Roboto:400,700,500italic,100italic|Roboto+Mono:300,400,700|EB+Garamond:400,400i');
@import url('https://fonts.googleapis.com/css?family=Roboto+Slab:100,400,700&display=swap');

.error_report {
	color:red; font-family:monospace;
}

div.output_wrapper {"""
    + body_font
    + """}

div.output_wrapper table,
div.jp-OutputArea-output table
{
	border-collapse:collapse;
}

div.output_wrapper table, div.output_wrapper th, div.output_wrapper td,
div.jp-OutputArea-output table, div.jp-OutputArea-output th, div.jp-OutputArea-output td
{
	border: 1px solid #999999;
	font-family:"Roboto Mono", monospace;
	font-size:9pt;
	font-weight:400;
}

div.output_wrapper th, div.output_wrapper td,
div.jp-OutputArea-output th, div.jp-OutputArea-output td
{
	padding:2px; text-align:left;
}

div.output_wrapper td.parameter_category,
div.jp-OutputArea-output td.parameter_category
{
	font-family:"Roboto", monospace;
	font-weight:500;
	background-color: #f4f4f4;
	font-style: italic;
}

div.output_wrapper th,
div.jp-OutputArea-output th
{
	font-family:"Roboto", monospace;
	font-weight:700;
}

div.output_wrapper table.dicta,
div.jp-OutputArea-output table.dicta
{
	border-left: 2px solid black; margin-bottom:2px; border-top:0; border-right:0
}

div.output_wrapper th.dicta, div.output_wrapper td.dicta,
div.jp-OutputArea-output th.dicta, div.jp-OutputArea-output td.dicta
{
	padding-top:0px; text-align:left; border:0;
}

div.output_wrapper div.LinearFunc,
div.jp-OutputArea-output div.LinearFunc
{
	font-family:"Roboto Mono", monospace;
	font-size:100%;
	font-weight:400;
}



.larch_signature {"""
    + signature_font
    + """ }
.larch_name_signature {"""
    + signature_name_font
    + """}

.larch_head_tag {font-size:150%; font-weight:900; font-family:"Roboto Slab", "Roboto", Verdana;}
.larch_head_tag_ver {font-size:80%; font-weight:200; font-family:"Roboto Slab", "Roboto", Verdana;}
.larch_head_tag_pth {
  font-size:40%; font-weight:200;
  font-family:"Roboto Slab", "Roboto", Verdana; padding-left:5px;
}
.larch_head_tag_more {
  font-size:50%; font-weight:300; font-family:"Roboto Mono", monospace; line-height:130%;
}

div.output_wrapper a.parameter_reference,
div.jp-OutputArea-output a.parameter_reference
{
	font-style: italic; text-decoration: none
}

div.output_wrapper .strut2,
div.jp-OutputArea-output .strut2
{
	min-width:1in
}

div.output_wrapper .histogram_cell,
div.jp-OutputArea-output .histogram_cell
{
	padding-top:1; padding-bottom:1; vertical-align:center;
}

div.output_wrapper .raw_log pre,
div.jp-OutputArea-output .raw_log pre
{
	font-family:"Roboto Mono", monospace;
	font-weight:300;
	font-size:70%;
}

.dicta pre
{
	margin:0;
}

div.output_wrapper caption,
div.jp-OutputArea-output caption,
{
    caption-side: bottom;
	text-align: left;
	font-family: Roboto;
	font-style: italic;
	font-weight: 100;
	font-size: 80%;
}

table.running_parameter_update caption
{
    caption-side: top;
	text-align: left;
	font-family: Roboto;
	font-style: italic;
	font-weight: 500;
	font-size: 100%;
}

table.dictionary
{
	border:0px hidden !important; border-collapse: collapse !important;
}

div.blurb {
	margin-top: 15px;
	max-width: 6.5in;
}

h2.figure_head {
	padding-left: .25in;
}

h3.figure_head {
	padding-left: .5in;
}

"""
)


deprecated_notebook_formatting = """

div.jp-RenderedMarkdown h1 {font-weight: 900; border-bottom:2px black solid; padding-bottom:4px}
div.jp-RenderedMarkdown h2 {font-weight: 850; border-bottom:0.5px black solid; padding-bottom:4px}
div.jp-RenderedMarkdown h3 {font-weight: 800; font-style:italic}

div.jp-RenderedMarkdown p {
    max-width:600px;
    font-size:150%;
    font-family: EB Garamond;
}

div.jp-RenderedMarkdown p code {
    font-size: 75%;
}


"""


css = HTML(f"<style>{_default_css_jupyter}\n\n{tooltipped_style().tostring()}</style>")


def stylesheet():
    display_html(css)
    # floatThead, floatTheadA = floating_table_head()
    # display_html(HTML(floatThead.tostring()))
    # display_html(HTML(floatTheadA.tostring()))



from collections.abc import MutableSequence

from .._optional import jax, jnp, js
from .param_core import ParameterBucket


class Mixture:
    def __init__(self):
        self._parent = None

    def param_names(self):
        """
        Named parameters referenced by this mixture, and their default values.

        Returns
        -------
        dict
        """
        raise NotImplementedError()

    def prep(self, bucket: ParameterBucket):
        raise NotImplementedError()

    def roll(self, draws: jax.Array, parameters: jax.Array) -> jax.Array:
        """
        Apply this mixing distribution to some random draws.

        Parameters
        ----------
        draws : jax.Array, shape [...]
            A set of pseudo-random draws, nominally uniformly distributed
            in the range 0 to 1.
        parameters : jax.Array, shape [..., n_params]
            An array of parameters, previously broadcasted to the same shape
            as the draws, plus the parameter dimension itself.

        Returns
        -------
        parameters : jax.Array, shape [..., n_params]
            The computed distribution of the target parameter has been overlaid.
        """
        raise NotImplementedError()

    def to_dict(self):
        raise NotImplementedError()


class MixtureList(MutableSequence):
    def __init__(self, init=None):
        self._parent = None
        self._mixtures = list()
        if init is not None:
            for i in init:
                if isinstance(i, Mixture):
                    self._mixtures.append(i)
                else:
                    raise TypeError(
                        f"members of {self.__class__.__name__} must be Mixture"
                    )

    def set_parent(self, instance):
        self._parent = instance
        for i in self._mixtures:
            i._parent = instance

    def __fresh(self, instance):
        newself = MixtureList()
        newself._instance = instance
        setattr(instance, self.private_name, newself)
        return newself

    def __mangle(self):
        try:
            self._parent.mangle()
        except AttributeError:
            pass

    def __get__(self, instance, owner):
        if instance is None:
            return self
        newself = getattr(instance, self.private_name, None)
        if newself is None:
            newself = self.__fresh(instance)
        return newself

    def __set__(self, instance, values):
        newself = getattr(instance, self.private_name, None)
        if newself is None:
            newself = self.__fresh(instance)
        else:
            newself._mixtures.clear()
        newself.__init__(values)
        newself.set_parent(instance)
        newself.__mangle()

    def __delete__(self, instance):
        newself = getattr(instance, self.private_name, None)
        if newself is not None and len(newself):
            newself.__mangle()
        if newself is None:
            newself = self.__fresh(instance)
        else:
            newself._mixtures.clear()
        newself.__init__()

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_private_" + name

    def __getitem__(self, item):
        return self._mixtures[item]

    def __setitem__(self, key: int, value):
        if not isinstance(value, Mixture):
            raise TypeError("items must be of type Mixture")
        self._mixtures[key] = value
        self.__mangle()
        self._mixtures[key]._parent = self._parent

    def __delitem__(self, key):
        del self._mixtures[key]
        self.__mangle()

    def __len__(self):
        return len(self._mixtures)

    def insert(self, index, value):
        if not isinstance(value, Mixture):
            raise TypeError("items must be of type Mixture")
        self._mixtures.insert(index, value)
        self.__mangle()
        self._mixtures[index]._parent = self._parent

    def __repr__(self):
        return repr(self._mixtures)

    def _is_duplicate(self, value):
        for i in self._mixtures:
            if i == value:
                return True
        return False

    def to_list(self):
        return [i.to_dict() for i in self._mixtures]

    def from_list(self, j):
        self._mixtures.clear()
        for i in j:
            kind = i.pop("type")
            if kind is None:
                raise ValueError("missing mixture type")
            cls = globals()[kind]
            self._mixtures.append(cls(**i))


class Normal(Mixture):
    def __init__(self, mean: str, std: str):
        super().__init__()
        self.mean_ = mean
        self.std_ = std
        self.imean = -1
        self.istd = -1
        self.default_mean = 0.0
        self.default_std = 0.001

    def __repr__(self):
        return f"{self.__class__.__name__}({self.mean_!r}, {self.std_!r})"

    def __eq__(self, other):
        return (
            isinstance(other, Normal)
            and self.mean_ == other.mean_
            and self.std_ == other.std_
        )

    def param_names(self):
        return {
            self.mean_: self.default_mean,
            self.std_: self.default_std,
        }

    def prep(self, bucket: ParameterBucket):
        self.imean = bucket.get_param_loc(self.mean_)
        self.istd = bucket.get_param_loc(self.std_)

    def roll(self, draw_vec, parameters):
        assert self.imean >= 0
        assert self.istd >= 0
        v = js.stats.norm.ppf(
            draw_vec, parameters[..., self.imean], parameters[..., self.istd]
        )
        parameters = parameters.at[..., self.imean].set(v)
        return parameters

    def to_dict(self):
        return dict(
            type=self.__class__.__name__,
            mean=self.mean_,
            std=self.std_,
        )


class LogNormal(Normal):
    def roll(self, draw_vec, parameters):
        assert self.imean >= 0
        assert self.istd >= 0
        v = js.stats.norm.ppf(
            draw_vec, parameters[..., self.imean], parameters[..., self.istd]
        )
        parameters = parameters.at[..., self.imean].set(jnp.exp(v))
        return parameters


class NegLogNormal(Normal):
    def roll(self, draw_vec, parameters):
        assert self.imean >= 0
        assert self.istd >= 0
        v = js.stats.norm.ppf(
            draw_vec, parameters[..., self.imean], parameters[..., self.istd]
        )
        parameters = parameters.at[..., self.imean].set(-jnp.exp(v))
        return parameters



#!/Users/jeffnewman/opt/anaconda3/envs/autumn22/bin/python3.1
import re
import sys

from jupyter_book.cli.main import main

if __name__ == "__main__":
    exec(open("_scripts/hide_test_cells.py").read())
    sys.argv[0] = re.sub(r"(-script\.pyw?|\.exe)?$", "", sys.argv[0])
    sys.exit(main())



import atexit
import os
import shutil
import tempfile
import types
import webbrowser

TemporaryBucket = []


def TemporaryBucketCleanUp():
    global TemporaryBucket
    for i in TemporaryBucket:
        try:
            os.remove(os.path.realpath(i.name))
        except PermissionError:
            try:
                shutil.rmtree(i.name)
            except Exception:
                pass
                # traceback.print_exc()
        except Exception:
            pass
            # traceback.print_exc()
    del TemporaryBucket


atexit.register(TemporaryBucketCleanUp)


def _open_in_chrome_or_something(url):
    # MacOS
    if os.path.exists("/Applications/Google Chrome.app"):
        chrome_path = r"open -a /Applications/Google\ Chrome.app %s"
    # Windows
    elif os.path.exists(r"C:\Program Files (x86)\Google\Chrome\Application\chrome.exe"):
        chrome_path = r"C:\Program Files (x86)\Google\Chrome\Application\chrome.exe %s"
    # Linux
    elif os.path.exists("/usr/bin/google-chrome"):
        chrome_path = "/usr/bin/google-chrome %s"
    else:
        chrome_path = None
    if chrome_path is None:
        webbrowser.open(url)
    else:
        webbrowser.get(chrome_path).open(url)


# Trap __exit__ to ensure the file does not get
# deleted when used in a with statement; to keep temporary files open
# until intentionally cleaned up
def _exit_without_closing(self, *arg, **kwarg):
    return self.file.__exit__(*arg, **kwarg)


def TemporaryFile(suffix="", mode="w+", use_chrome=True):
    t = tempfile.NamedTemporaryFile(suffix=suffix, mode=mode, delete=False)
    t.__exit__ = types.MethodType(_exit_without_closing, t)
    if use_chrome:
        t.view = types.MethodType(
            lambda self: _open_in_chrome_or_something(
                "file://" + os.path.realpath(self.name)
            ),
            t,
        )
    else:
        t.view = types.MethodType(
            lambda self: webbrowser.open("file://" + os.path.realpath(self.name)), t
        )
    global TemporaryBucket
    TemporaryBucket.append(t)
    return t


def _try_write(self, content):
    try:
        self.write_(content)
    except Exception:
        try:
            self.write_(str(content))
        except Exception:
            self.write_(str(content).encode("utf-8"))


def TemporaryHtml(style=None, *, nohead=False, mode="wb+", content=None, **tagheads):
    t = TemporaryFile(suffix=".html", mode=mode)
    if "b" in mode:
        t.write_ = t.write
        t.write = lambda x: _try_write(t, x)
    if not nohead and (style or len(tagheads) > 0):
        t.write("<head>")
        if style:
            t.write(f"<style>{style}</style>")
        for tag, content in tagheads.items():
            t.write(f"<{tag.lower()}>{content}</{tag.lower()}>")
        t.write("</head>")
    if content is not None:
        t.write(content)
        t.view()
    return t


def TemporaryGzipInflation(gzfile):
    t = TemporaryFile(mode="wb")
    import gzip

    with gzip.open(gzfile, "rb") as previewfile:
        t.write(previewfile.read())
    t.flush()
    return t.name


def TemporaryZipInflation(zfile):
    t = TemporaryDirectory(common=False)
    from zipfile import ZipFile

    ZipFile(zfile).extractall(path=t)
    return t


_tempdir = None


def TemporaryDirectory(common=True):
    global _tempdir
    if common and _tempdir is not None:
        return _tempdir
    t = tempfile.mkdtemp()
    global TemporaryBucket
    from . import Dict

    TemporaryBucket.append(Dict(name=t))
    if common:
        _tempdir = t
    return t


def TemporaryCopy(sourcefile, spool=True):
    tdir = TemporaryDirectory()
    basename = os.path.basename(sourcefile)
    if spool:
        n = 0
        vbasename = basename
        while os.path.exists(os.path.join(tdir, vbasename)):
            n += 1
            basename_ = os.path.splitext(basename)
            vbasename = f"{basename_[0]}.{n}{basename_[1]}"
        shutil.copy2(sourcefile, os.path.join(tdir, vbasename))
        return os.path.join(tdir, vbasename)
    else:
        shutil.copy2(sourcefile, tdir)
        return os.path.join(tdir, os.path.basename(sourcefile))



import copy

import pandas as pd
from bs4 import BeautifulSoup


class html_tables:
    def __init__(self, raw_html):
        self.url_soup = BeautifulSoup(raw_html, "lxml")

    def read(self):
        self.tables = []
        self.tables_html = self.url_soup.find_all("table")

        # Parse each table
        for n in range(0, len(self.tables_html)):
            n_cols = 0
            n_rows = 0

            for row in self.tables_html[n].find_all("tr"):
                col_tags = row.find_all(["td", "th"])
                if len(col_tags) > 0:
                    n_rows += 1
                    if len(col_tags) > n_cols:
                        n_cols = len(col_tags)

            # Create dataframe
            df = pd.DataFrame(index=range(0, n_rows), columns=range(0, n_cols))

            # Create list to store rowspan values
            skip_index = [0 for i in range(0, n_cols)]
            this_skip_index = copy.deepcopy(skip_index)

            # Start by iterating over each row in this table...
            row_counter = 0
            for row in self.tables_html[n].find_all("tr"):
                # Skip row if it's blank
                if len(row.find_all(["td", "th"])) == 0:
                    pass

                else:
                    # Get all cells containing data in this row
                    columns = row.find_all(["td", "th"])
                    col_dim = []
                    row_dim = []
                    col_dim_counter = -1
                    row_dim_counter = -1
                    col_counter = -1
                    this_skip_index = copy.deepcopy(skip_index)

                    for col in columns:
                        # Determine cell dimensions
                        colspan = col.get("colspan")
                        if colspan is None:
                            col_dim.append(1)
                        else:
                            col_dim.append(int(colspan))
                        col_dim_counter += 1

                        rowspan = col.get("rowspan")
                        if rowspan is None:
                            row_dim.append(1)
                        else:
                            row_dim.append(int(rowspan))
                        row_dim_counter += 1

                        # Adjust column counter
                        if col_counter == -1:
                            col_counter = 0
                        else:
                            col_counter = col_counter + col_dim[col_dim_counter - 1]

                        try:
                            while skip_index[col_counter] > 0:
                                col_counter += 1
                        except IndexError:
                            from pprint import pprint

                            print("~" * 50)
                            pprint(locals())
                            print("~" * 50)
                            raise
                            pass

                        # Get cell contents
                        cell_data = col.get_text()

                        # Insert data into cell
                        df.iat[row_counter, col_counter] = cell_data

                        # Record column skipping index
                        if row_dim[row_dim_counter] > 1:
                            this_skip_index[col_counter] = row_dim[row_dim_counter]

                # Adjust row counter
                row_counter += 1

                # Adjust column skipping index
                skip_index = [i - 1 if i > 0 else i for i in this_skip_index]

            # Append dataframe to list of tables
            self.tables.append(df)

        return self.tables


def xml_table_to_txt(elem):
    first_table = html_tables(elem.tostring()).read()[0]
    return first_table.to_string(
        buf=None,
        columns=None,
        col_space=None,
        header=True,
        index=False,
        na_rep="",
        formatters=None,
        float_format=None,
        sparsify=None,
        index_names=False,
        justify=None,
        line_width=None,
        max_rows=None,
        max_cols=None,
        show_dimensions=False,
    ).partition("\n")[2]



###############################################################################
# Auto-generated by `jupyter-book config`
# If you wish to continue using _config.yml, make edits to that file and
# re-generate this one.
###############################################################################

import larch

html_static_path = ["_static"]

html_css_files = [
    "css/custom.css",
]

add_module_names = False
author = "The Jupyter Book community"
autodoc_default_options = {"show-inheritance": True}
autosummary_generate = True
bibtex_bibfiles = ["references.bib"]
bibtex_reference_style = "author_year"
comments_config = {"hypothesis": False, "utterances": False}
copyright = "2024"
exclude_patterns = ["**.ipynb_checkpoints", ".DS_Store", "Thumbs.db", "_build"]
extensions = [
    "sphinx_togglebutton",
    "sphinx_copybutton",
    "myst_nb",
    # "jupyter_book",
    "pydata_sphinx_theme",
    "sphinx_thebe",
    "sphinx_comments",
    # "sphinx_external_toc",
    "sphinx.ext.intersphinx",
    "sphinx_design",
    # "sphinx_book_theme",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx.ext.doctest",
    "sphinx.ext.graphviz",
    "sphinx.ext.viewcode",
    "sphinxcontrib.bibtex",
    # "sphinx_jupyterbook_latex",
]
# external_toc_exclude_missing = False
# external_toc_path = "_toc.yml"
html_baseurl = ""
html_favicon = "_img/larch_favicon.png"
# html_logo = "_img/larch-logo.png"
html_show_copyright = False
html_sourcelink_suffix = ""
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "search_bar_text": "Search these docs...",
    "favicons": [{"rel": "icon", "sizes": "32x32", "href": "_img/larch_favicon.png"}],
    "navigation_with_keys": True,
    "logo": {
        "text": f"v{larch.__version__}",
        "image_light": "_img/larch-logo.png",
        "image_dark": "_img/larch-logo.png",
    },
}
html_title = f"v{larch.__version__}"
latex_engine = "pdflatex"
myst_enable_extensions = [
    "colon_fence",
    "dollarmath",
    "linkify",
    "substitution",
    "tasklist",
]
myst_url_schemes = ["mailto", "http", "https"]
napoleon_use_rtype = False
nb_execution_allow_errors = True
nb_execution_cache_path = ""
nb_execution_excludepatterns = []
nb_execution_in_temp = False
nb_execution_mode = "cache"
nb_execution_timeout = 300
nb_output_stderr = "show"
numfig = True
pygments_style = "sphinx"
rst_epilog = (
    ".. |idca| replace:: :ref:`idca <idca>`\n.. |idco| replace:: :ref:`idco <idco>`\n"
)
suppress_warnings = ["myst.domains"]
use_jupyterbook_latex = True
use_multitoc_numbering = True



from pathlib import Path

import numpy as np
import pandas as pd

from ... import DataFrames, Model, P
from .. import Dict
from .general import (
    apply_coefficients,
    clean_values,
    construct_nesting_tree,
    explicit_value_parameters,
    linear_utility_from_spec,
    simple_simulate_data,
)


def tour_mode_choice_model(
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    data = simple_simulate_data(
        name="tour_mode_choice",
        edb_directory=edb_directory,
    )
    coefficients = data.coefficients
    coef_template = data.coef_template
    spec = data.spec
    chooser_data = data.chooser_data
    settings = data.settings

    chooser_data = clean_values(
        chooser_data,
        data.alt_names,
        alt_names_to_codes=data.alt_names_to_codes,
        choice_code="override_choice_code",
    )

    tree = construct_nesting_tree(data.alt_names, settings["NESTS"])

    purposes = list(coef_template.columns)

    # Setup purpose specific models
    m = {purpose: Model(graph=tree) for purpose in purposes}
    for alt_code, alt_name in tree.elemental_names().items():
        # Read in base utility function for this alt_name
        u = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col=alt_name,
            ignore_x=("#",),
        )
        for purpose in purposes:
            # Modify utility function based on template for purpose
            u_purp = sum(
                (P(coef_template[purpose].get(i.param, i.param)) * i.data * i.scale)
                for i in u
            )
            m[purpose].utility_co[alt_code] = u_purp

    for model in m.values():
        explicit_value_parameters(model)
    apply_coefficients(coefficients, m)

    avail = {}
    for acode, _aname in data.alt_codes_to_names.items():
        unavail_cols = list(
            (
                chooser_data[i.data]
                if i.data in chooser_data
                else chooser_data.eval(i.data)
            )
            for i in m[purposes[0]].utility_co[acode]
            if i.param == "-999"
        )
        if len(unavail_cols):
            avail[acode] = sum(unavail_cols) == 0
        else:
            avail[acode] = 1
    avail = pd.DataFrame(avail).astype(np.int8)
    avail.index = chooser_data.index

    d = DataFrames(
        co=chooser_data,
        av=avail,
        alt_codes=data.alt_codes,
        alt_names=data.alt_names,
    )

    for purpose, model in m.items():
        model.dataservice = d.selector_co(f"tour_type=='{purpose}'")
        model.choice_co_code = "override_choice_code"

    from larch.model.model_group import ModelGroup

    mg = ModelGroup(m.values())

    if return_data:
        return mg, Dict(
            edb_directory=Path(edb_directory),
            chooser_data=chooser_data,
            avail=avail,
            coefficients=coefficients,
            coef_template=coef_template,
            spec=spec,
        )

    return mg



import os
import warnings
from collections.abc import Mapping

import numpy
import pandas
import tables as _tb

from .util import Dict
from .util.text_manip import truncate_path_for_display

if os.environ.get("READTHEDOCS", None) == "True":
    _omx_base_class = object
else:
    _omx_base_class = _tb.file.File


class OMX_Error(Exception):
    pass


class OMXBadFormat(OMX_Error):
    pass


class OMXIncompatibleShape(OMXBadFormat):
    pass


class OMXNonUniqueLookup(OMX_Error):
    pass


class OMX(_omx_base_class):
    """A subclass of the :class:`tables.File` class, adding an interface for openmatrix files.

    As suggested in the openmatrix documentation, the default when creating an OMX file
    is to use zlib compression level 1, although this can be overridden.
    """

    def __repr__(self):
        from .util.text_manip import max_len

        s = f"<larch.OMX> {truncate_path_for_display(self.filename)}"
        s += f"\n |  shape:{self.shape}"
        if len(self.data._v_children):
            s += "\n |  data:"
        just = max_len(self.data._v_children.keys())
        for i in sorted(self.data._v_children.keys()):
            s += "\n |    {0:{2}s} ({1})".format(
                i, self.data._v_children[i].dtype, just
            )
        if len(self.lookup._v_children):
            s += "\n |  lookup:"
        just = max_len(self.lookup._v_children.keys())
        for i in sorted(self.lookup._v_children.keys()):
            s += "\n |    {0:{3}s} ({1} {2})".format(
                i,
                self.lookup._v_children[i].shape[0],
                self.lookup._v_children[i].dtype,
                just,
            )
        return s

    def __init__(self, *arg, complevel=1, complib="zlib", **kwarg):
        if len(arg) > 0 and isinstance(arg[0], str) and arg[0][-3:] == ".gz":
            from .util.temporaryfile import TemporaryGzipInflation

            self._temp_inflated_filename = TemporaryGzipInflation(arg[0])
            arg = (self._temp_inflated_filename,) + arg[1:]
            if "mode" in kwarg and kwarg["mode"] != "r":
                raise TypeError("cannot open .gz with other than readonly mode")
            if len(arg) > 1 and arg[1] != "r":
                raise TypeError("cannot open .gz with other than readonly mode")

        if len(arg) == 0:
            # Make a temporary file
            from .util.temporaryfile import TemporaryDirectory

            self._temporary_dir = TemporaryDirectory()
            kwarg["driver"] = "H5FD_CORE"
            kwarg["driver_core_backing_store"] = 0
            n = 1
            while os.path.exists(os.path.join(self._temporary_dir, f"TEMP{n}.omx")):
                n += 1
            arg = (os.path.join(self._temporary_dir, f"TEMP{n}.omx"), "w")

        if "filters" in kwarg:
            super().__init__(*arg, **kwarg)
        else:
            super().__init__(
                *arg, filters=_tb.Filters(complib=complib, complevel=complevel), **kwarg
            )
        try:
            self.data = self._get_or_create_path("/data", True)
        except _tb.exceptions.FileModeError:
            raise OMXBadFormat(
                "the '/data' node does not exist and cannot be created"
            ) from None
        try:
            self.lookup = self._get_or_create_path("/lookup", True)
        except _tb.exceptions.FileModeError:
            raise OMXBadFormat(
                "the '/lookup' node does not exist and cannot be created"
            ) from None
        if "OMX_VERSION" not in self.root._v_attrs:
            try:
                self.root._v_attrs.OMX_VERSION = b"0.2"
            except _tb.exceptions.FileModeError:
                raise OMXBadFormat(
                    "the root OMX_VERSION attribute does not exist and cannot be created"
                ) from None
        if "SHAPE" not in self.root._v_attrs:
            try:
                self.root._v_attrs.SHAPE = numpy.zeros(2, dtype=int)
            except _tb.exceptions.FileModeError:
                raise OMXBadFormat(
                    "the root SHAPE attribute does not exist and cannot be created"
                ) from None
        self.rlookup = Dict()
        self.rlookup._helper = self.get_reverse_lookup

    @property
    def version(self):
        """The open matrix format version for this file."""
        if "OMX_VERSION" not in self.root._v_attrs:
            raise OMXBadFormat("the root OMX_VERSION attribute does not exist")
        return self.root._v_attrs.OMX_VERSION

    def change_mode(self, mode, **kwarg):
        """
        Change the file mode of the underlying HDF5 file.

        Can be used to change from read-only to read-write.

        Parameters
        ----------
        mode : {'r','a'}
            The file mode to set. Note that the 'w' mode, which would
            open a blank HDF5 overwriting an existing file, is not allowed
            here.

        Returns
        -------
        self
        """
        try:
            if mode == self.mode:
                return
            if mode == "w":
                raise TypeError("cannot change_mode to w, close the file and delete it")
        except RecursionError:
            pass  # the file was probably closed elsewhere, reopen it
        filename = self.filename
        self.close()
        self.__init__(filename, mode, **kwarg)
        return self

    def keys(self, kind="all"):
        """
        Get the names of data and/or lookup stored in this file.

        Parameters
        ----------
        kind : {'all', 'data', 'lookup'}

        Returns
        -------
        list
        """
        if kind == "data":
            return list(self.data._v_children.keys())
        elif kind == "lookup":
            return list(self.lookup._v_children.keys())
        else:
            return list(self.data._v_children.keys()) + list(
                self.lookup._v_children.keys()
            )

    @property
    def shape(self):
        """The shape of the OMX file.

        As required by the standard, all OMX files must have a two dimensional shape. This
        attribute accesses or alters that shape. Note that attempting to change the
        shape of an existing file that already has data tables that would be incompatible
        with the new shape will raise an OMXIncompatibleShape exception.
        """
        sh = self.root._v_attrs.SHAPE[:]
        proposal = (sh[0], sh[1])
        if proposal == (0, 0) and self.data._v_nchildren > 0:
            first_child_name = next(iter(self.data._v_children.keys()))
            proposal = self.data._v_children[first_child_name].shape
            shp = numpy.empty(2, dtype=int)
            shp[0] = proposal[0]
            shp[1] = proposal[1]
            self.root._v_attrs.SHAPE = shp
        return proposal

    @shape.setter
    def shape(self, x):
        if self.data._v_nchildren > 0:
            if x[0] != self.shape[0] and x[1] != self.shape[1]:
                raise OMXIncompatibleShape(
                    f"this omx has shape {self.shape!s} but you want to set {x!s}"
                )
        if self.data._v_nchildren == 0:
            shp = numpy.empty(2, dtype=int)
            shp[0] = x[0]
            shp[1] = x[1]
            self.root._v_attrs.SHAPE = shp

    def crop(self, shape0, shape1=None, *, complevel=1, complib="zlib"):
        if isinstance(shape0, (tuple, list)) and len(shape0) == 2 and shape1 is None:
            shape0, shape1 = shape0[0], shape0[1]
        elif isinstance(shape0, (int)) and shape1 is None:
            shape1 = shape0
        start0, start1 = self.shape[0], self.shape[1]
        if start0 == shape0 and start1 == shape1:
            return
        if shape0 > start0:
            raise TypeError(
                f"crop must shrink the shape, but {shape0} > {start0} in first dimension"
            )
        if shape1 > start1:
            raise TypeError(
                f"crop must shrink the shape, but {shape1} > {start1} in first dimension"
            )
        if shape0 != shape1 and start0 == start1:
            raise TypeError(
                "do not automatically crop a square array into a non-square shape, "
                "results may be ambiguous"
            )
        for i in self.data._v_children:
            self.add_matrix(
                i,
                self.data._v_children[i][:shape0, :shape1],
                overwrite=True,
                complevel=complevel,
                complib=complib,
                ignore_shape=True,
            )
        for i in self.lookup._v_children:
            if self.lookup[i].shape[0] == start0:
                self.add_lookup(
                    i,
                    self.lookup._v_children[i][:shape0],
                    overwrite=True,
                    complevel=complevel,
                    complib=complib,
                    ignore_shape=True,
                )
            elif self.lookup[i].shape[1] == start1:
                self.add_lookup(
                    i,
                    self.lookup._v_children[i][:shape1],
                    overwrite=True,
                    complevel=complevel,
                    complib=complib,
                    ignore_shape=True,
                )
        shp = numpy.empty(2, dtype=int)
        shp[0] = shape0
        shp[1] = shape1
        self.root._v_attrs.SHAPE = shp

    def add_blank_lookup(
        self, name, atom=None, shape=None, complevel=1, complib="zlib", **kwargs
    ):
        """
        Add a blank lookup to the OMX file.

        Parameters
        ----------
        name : str
            The name of the matrix to add.
        atom : tables.Atom, optional
            The atomic data type for the new matrix. If not given, Float64 is assumed.
        shape : int, optional
            The length of the lookup to add. Must match one of the dimensions of an existing
            shape.  If there is no existing shape, the shape will be initialized as a square
            with this size.
        complevel : int, default 1
            Compression level.
        complib : str, default 'zlib'
            Compression library.

        Returns
        -------
        tables.CArray

        Note
        ----
        This method allows you to add a blank matrix of 3 or more dimemsions,
        only the first 2 must match the OMX.  Adding a matrix of more than
        two dimensions may violate compatability with other open matrix tools,
        do so at your own risk.
        """
        if name in self.lookup._v_children:
            return self.lookup._v_children[name]
        if atom is None:
            atom = _tb.Float64Atom()
        if self.shape == (0, 0) and shape is None:
            raise OMXBadFormat("must set a nonzero shape first or give a shape")
        if shape is not None:
            if shape != self.shape[0] and shape != self.shape[1]:
                if self.shape[0] == 0 and self.shape[1] == 0:
                    self.shape = (shape, shape)
                else:
                    raise OMXIncompatibleShape(
                        f"this omx has shape {self.shape!s} but you want to set {shape!s}"
                    )
        else:
            if self.shape[0] != self.shape[1]:
                raise OMXIncompatibleShape(
                    f"this omx has shape {self.shape!s} but you did not pick one"
                )
            shape = self.shape[0]
        return self.create_carray(
            self.lookup,
            name,
            atom=atom,
            shape=numpy.atleast_1d(shape),
            filters=_tb.Filters(complib=complib, complevel=complevel),
            **kwargs,
        )

    def add_blank_matrix(
        self, name, atom=None, shape=None, complevel=1, complib="zlib", **kwargs
    ):
        """
        Add a blank matrix to the OMX file.

        Parameters
        ----------
        name : str
            The name of the matrix to add.
        atom : tables.Atom, optional
            The atomic data type for the new matrix. If not given, Float64 is assumed.
        shape : tuple, optional
            The shape of the matrix to add.  The first two dimensions must match the
            shape of any existing matrices; giving a shape is mostly used on initialization.
        complevel : int, default 1
            Compression level.
        complib : str, default 'zlib'
            Compression library.

        Returns
        -------
        tables.CArray

        Note
        ----
        This method allows you to add a blank matrix of 3 or more dimemsions,
        only the first 2 must match the OMX.  Adding a matrix of more than
        two dimensions may violate compatability with other open matrix tools,
        do so at your own risk.
        """
        if name in self.data._v_children:
            return self.data._v_children[name]
        if atom is None:
            atom = _tb.Float64Atom()
        if self.shape == (0, 0):
            raise OMXBadFormat("must set a nonzero shape first")
        if shape is not None:
            if shape[:2] != self.shape:
                raise OMXIncompatibleShape(
                    f"this omx has shape {self.shape!s} but you want to set {shape[:2]!s}"
                )
        else:
            shape = self.shape
        return self.create_carray(
            self.data,
            name,
            atom=atom,
            shape=shape,
            filters=_tb.Filters(complib=complib, complevel=complevel),
            **kwargs,
        )

    def add_matrix(
        self,
        name,
        obj,
        *,
        overwrite=False,
        complevel=1,
        complib="zlib",
        ignore_shape=False,
        **kwargs,
    ):
        """
        Add a matrix to the OMX file.

        Parameters
        ----------
        name : str
            The name of the matrix to add.
        obj : array-like
            The data for the new matrix.
        overwrite : bool, default False
            Whether to overwrite an existing matrix with the same name.
        complevel : int, default 1
            Compression level.
        complib : str, default 'zlib'
            Compression library.
        ignore_shape : bool, default False
            Whether to ignore all checks on the shape of the matrix being added.
            Setting this to True allows adding a matrix with a different shape
            than any other existing matrices, which may result in an invalid OMX
            file.

        Returns
        -------
        tables.CArray
        """
        obj = numpy.asanyarray(obj)
        if not ignore_shape:
            if len(obj.shape) != 2:
                raise OMXIncompatibleShape(
                    "all omx arrays must have 2 dimensional shape"
                )
            if self.data._v_nchildren > 0:
                if obj.shape != self.shape:
                    raise OMXIncompatibleShape(
                        f"this omx has shape {self.shape!s} but you want to add {obj.shape!s}"
                    )
        if self.data._v_nchildren == 0:
            shp = numpy.empty(2, dtype=int)
            shp[0] = obj.shape[0]
            shp[1] = obj.shape[1]
            self.root._v_attrs.SHAPE = shp
        if name in self.data._v_children and not overwrite:
            raise TypeError(f"{name} exists")
        if name in self.data._v_children:
            self.remove_node(self.data, name)
        return self.create_carray(
            self.data,
            name,
            obj=obj,
            filters=_tb.Filters(complib=complib, complevel=complevel),
            **kwargs,
        )

    def add_lookup(
        self,
        name,
        obj,
        *,
        overwrite=False,
        complevel=1,
        complib="zlib",
        ignore_shape=False,
        **kwargs,
    ):
        """
        Add a lookup mapping to the OMX file.

        Parameters
        ----------
        name : str
            The name of the matrix to add.
        obj : array-like
            The data for the new matrix.
        overwrite : bool, default False
            Whether to overwrite an existing matrix with the same name.
        complevel : int, default 1
            Compression level.
        complib : str, default 'zlib'
            Compression library.
        ignore_shape : bool, default False
            Whether to ignore all checks on the shape of the lookup being added.
            Setting this to True allows adding a lookup with a different length
            than either of the two shape dimensions, which may result in an invalid
            OMX file.

        Returns
        -------
        tables.CArray
        """
        obj = numpy.asanyarray(obj)
        if not ignore_shape:
            if len(obj.shape) != 1:
                raise OMXIncompatibleShape(
                    "all omx lookups must have 1 dimensional shape"
                )
            if self.data._v_nchildren > 0:
                if obj.shape[0] not in self.shape:
                    raise OMXIncompatibleShape(
                        "this omx has shape {!s} but you want to add a lookup with {!s}".format(
                            self.shape, obj.shape
                        )
                    )
        if self.data._v_nchildren == 0 and self.shape == (0, 0):
            raise OMXIncompatibleShape(
                "don't add lookup to omx with no data and no shape"
            )
        if name in self.lookup._v_children and not overwrite:
            raise TypeError(f"{name} exists")
        if name in self.lookup._v_children:
            self.remove_node(self.lookup, name)
        return self.create_carray(
            self.lookup,
            name,
            obj=obj,
            filters=_tb.Filters(complib=complib, complevel=complevel),
            **kwargs,
        )

    def change_all_atoms_of_type(self, oldatom, newatom):
        for name in self.data._v_children:
            if self.data._v_children[name].dtype == oldatom:
                print(f"changing matrix {name} from {oldatom} to {newatom}")
                self.change_atom_type(name, newatom, matrix=True, lookup=False)
        for name in self.lookup._v_children:
            if self.lookup._v_children[name].dtype == oldatom:
                print(f"changing lookup {name} from {oldatom} to {newatom}")
                self.change_atom_type(name, newatom, matrix=False, lookup=True)

    def change_atom_type(
        self, name, atom, matrix=True, lookup=True, require_smaller=True
    ):
        if isinstance(atom, numpy.dtype):
            atom = _tb.Atom.from_dtype(atom)
        elif not isinstance(atom, _tb.atom.Atom):
            atom = _tb.Atom.from_type(atom)
        if matrix:
            if name in self.data._v_children:
                orig = self.data._v_children[name]
                neww = self.add_blank_matrix(name + "_temp_atom", atom=atom)
                for i in range(self.shape[0]):
                    neww[i] = orig[i]
                if require_smaller and neww.size_on_disk >= orig.size_on_disk:
                    warnings.warn(
                        f"abort change_atom_type on {name}, {neww.size_on_disk} > {orig.size_on_disk}",
                        stacklevel=2,
                    )
                    neww._f_remove()
                else:
                    neww._f_rename(name, overwrite=True)
        if lookup:
            if name in self.lookup._v_children:
                orig = self.lookup._v_children[name]
                neww = self.add_blank_lookup(name + "_temp_atom", atom=atom)
                for i in range(self.shape[0]):
                    neww[i] = orig[i]
                if require_smaller and neww.size_on_disk >= orig.size_on_disk:
                    warnings.warn(
                        f"abort change_atom_type on {name}, {neww.size_on_disk} > {orig.size_on_disk}",
                        stacklevel=2,
                    )
                    neww._f_remove()
                else:
                    neww._f_rename(name, overwrite=True)

    def get_reverse_lookup(self, name):
        labels = self.lookup._v_children[name][:]
        label_to_i = dict(enumerate(labels))
        self.rlookup[name] = label_to_i
        return label_to_i

    def lookup_to_index(self, lookupname, arr):
        """
        Convert an array of lookup-able values into indexes.

        If you have an array of lookup-able values (e.g., TAZ identifiers) and you
        want to convert them to 0-based indexes for use in accessing matrix data,
        this is the function for you.

        Parameters
        ----------
        lookupname : str
            The name of the lookup in the open matrix file. This lookup must already
            exist and have a set of unique lookup values.
        arr : array-like
            An array of values that appear in the lookup. This method uses
            numpy.digitize to process values, so any target value that appears in `arr` but does
            not appear in the lookup will be assigned the index of the smallest lookup
            value that is greater than the target, or the maximum lookup value if no lookup value
            is greater than the target.

        Returns
        -------
        array
            An array of index (int) values, with the same shape as `arr`.

        Raises
        ------
        OMXNonUniqueLookup
            When the lookup does not contain a set of unique values, this tool is not appropriate.

        """
        from .util.arraytools import is_sorted_and_unique

        labels = self.lookup._v_children[lookupname][:]
        if is_sorted_and_unique(labels):
            return numpy.digitize(arr, labels, right=True)
        uniq_labels, uniq_indexes = numpy.unique(labels, return_inverse=True)
        if len(uniq_labels) != len(labels):
            raise OMXNonUniqueLookup(
                f"lookup '{lookupname}' does not have unique labels for each item"
            )
        index_malordered = numpy.digitize(arr, uniq_labels, right=True)
        return uniq_indexes[index_malordered]

    def import_datatable(
        self,
        filepath,
        one_based=True,
        chunksize=10000,
        column_map=None,
        default_atom="float32",
        log=None,
    ):
        """Import a table in r,c,x,x,x... format into the matrix.

        The r and c columns need to be either 0-based or 1-based index values
        (this may be relaxed in the future). The matrix must already be set up
        with the correct size before importing the datatable.

        Parameters
        ----------
        filepath : str or buffer
            This argument will be fed directly to the :func:`pandas.read_csv` function.
        one_based : bool
            If True (the default) it is assumed that zones are indexed sequentially
            starting with 1 (as is typical for travel demand forecasting applications).
            Otherwise, it is assumed that zones are indexed sequentially starting
            with 0 (typical for other c and python applications).
        chunksize : int
            The number of rows of the source file to read as a chunk.  Reading a
            giant file in moderate sized chunks can be much faster and less memory
            intensive than reading the entire file.
        column_map : dict or None
            If given, this dict maps columns of the input file to OMX tables, with
            the keys as the columns in the input and the values as the tables in
            the output.
        default_atom : str or dtype
            The default atomic type for imported data when the table does not already
            exist in this openmatrix.
        """
        if log is not None:
            log("START import_datatable")
        from .util.smartread import SmartFileReader

        sfr = SmartFileReader(filepath)
        reader = pandas.read_csv(sfr, chunksize=chunksize)
        chunk0 = next(pandas.read_csv(filepath, chunksize=10))
        offset = 1 if one_based else 0

        if column_map is None:
            column_map = {i: i for i in chunk0.columns}

        if isinstance(default_atom, str):
            default_atom = _tb.Atom.from_dtype(numpy.dtype(default_atom))
        elif isinstance(default_atom, numpy.dtype):
            default_atom = _tb.Atom.from_dtype(default_atom)

        for t in column_map.values():
            self.add_blank_matrix(t, atom=default_atom)

        for n, chunk in enumerate(reader):
            if log is not None:
                log(f"PROCESSING CHUNK {n} [{sfr.progress()}]")
            r = chunk.iloc[:, 0].values - offset
            c = chunk.iloc[:, 1].values - offset

            if not numpy.issubdtype(r.dtype, numpy.integer):
                r = r.astype(int)
            if not numpy.issubdtype(c.dtype, numpy.integer):
                c = c.astype(int)

            for col in chunk.columns:
                if col in column_map:
                    self.data._v_children[column_map[col]][r, c] = chunk[col].values
                    self.data._v_children[column_map[col]].flush()
            if log is not None:
                log(f"finished processing chunk {n} [{sfr.progress()}]")

            self.flush()

        if log is not None:
            log(f"import_datatable({filepath}) complete")

    def import_datatable_3d(
        self,
        filepath,
        one_based=True,
        chunksize=10000,
        default_atom="float32",
        log=None,
    ):
        """Import a table in r,c,x,x,x... format into the matrix.

        The r and c columns need to be either 0-based or 1-based index values
        (this may be relaxed in the future). The matrix must already be set up
        with the correct size before importing the datatable.

        This method is functionally the same as :meth:`import_datatable` but uses
        a different implementation. It is much more memory intensive but also much
        faster than the non-3d version.

        Parameters
        ----------
        filepath : str or buffer
            This argument will be fed directly to the :func:`pandas.read_csv` function.
        one_based : bool
            If True (the default) it is assumed that zones are indexed sequentially
            starting with 1 (as is typical for travel demand forecasting applications).
            Otherwise, it is assumed that zones are indexed sequentially starting
            with 0 (typical for other c and python applications).
        chunksize : int
            The number of rows of the source file to read as a chunk.  Reading a
            giant file in moderate sized chunks can be much faster and less memory
            intensive than reading the entire file.
        default_atom : str or dtype
            The default atomic type for imported data when the table does not already
            exist in this openmatrix.
        """
        if log is not None:
            log("START import_datatable")
        from .util.smartread import SmartFileReader

        sfr = SmartFileReader(filepath)
        reader = pandas.read_csv(sfr, chunksize=chunksize)
        chunk0 = next(pandas.read_csv(filepath, chunksize=10))
        offset = 1 if one_based else 0

        if isinstance(default_atom, str):
            default_dtype = numpy.dtype(default_atom)
            default_atom = _tb.Atom.from_dtype(numpy.dtype(default_atom))
        elif isinstance(default_atom, numpy.dtype):
            default_dtype = default_atom
            default_atom = _tb.Atom.from_dtype(default_atom)
        else:
            default_dtype = default_atom.dtype

        # self.add_blank_matrix(matrixname, atom=default_atom, shape=self.shape+(len(chunk0.columns)-2,))

        temp_slug = numpy.zeros(
            self.shape + (len(chunk0.columns) - 2,), dtype=default_dtype
        )

        for n, chunk in enumerate(reader):
            if log is not None:
                log(f"PROCESSING CHUNK {n} [{sfr.progress()}]")
            r = chunk.iloc[:, 0].values - offset
            c = chunk.iloc[:, 1].values - offset

            if not numpy.issubdtype(r.dtype, numpy.integer):
                r = r.astype(int)
            if not numpy.issubdtype(c.dtype, numpy.integer):
                c = c.astype(int)

            temp_slug[r, c] = chunk.values[:, 2:]
            if log is not None:
                log(f"finished processing chunk {n} [{sfr.progress()}]")

            self.flush()

        for cn, colname in enumerate(chunk0.columns[2:]):
            self.add_blank_matrix(colname, atom=default_atom, shape=self.shape)
            self.data._v_children[colname][:] = temp_slug[:, :, cn]

        if log is not None:
            log(f"import_datatable({filepath}) complete")

    @classmethod
    def import_dbf(cls, dbffile, omxfile, shape, o, d, cols, smallest_zone_number=1):
        try:
            from simpledbf import Dbf5
        except ImportError:
            raise
        dbf = Dbf5(dbffile, codec="utf-8")
        tempstore = {c: numpy.zeros(shape, dtype=numpy.float32) for c in cols}
        for df in dbf.to_dataframe(chunksize=shape[1]):
            oz = df[o].values.astype(int) - smallest_zone_number
            dz = df[d].values.astype(int) - smallest_zone_number
            for c in cols:
                tempstore[c][oz, dz] = df[c].values
        omx = cls(omxfile, mode="a")
        for c in cols:
            omx.add_matrix(c, tempstore[c])
        omx.flush()
        return omx

    def __getitem__(self, key):
        if isinstance(key, str):
            if key in self.data._v_children:
                if key in self.lookup._v_children:
                    raise KeyError(f"key {key} found in both data and lookup")
                else:
                    return self.data._v_children[key]
            if key in self.lookup._v_children:
                return self.lookup._v_children[key]
            raise KeyError(f"matrix named {key} not found")
        raise TypeError("OMX matrix access must be by name (str)")

    def __setitem__(self, key, value):
        try:
            value_shape = value.shape
        except Exception as err:
            raise TypeError("value must array-like with one or two dimensions") from err
        if len(value_shape) == 1:
            if value_shape[0] == self.shape[0] or value_shape[0] == self.shape[1]:
                self.add_lookup(key, numpy.asarray(value))
            else:
                raise OMXIncompatibleShape(
                    f"cannot add vector[{value_shape[0]}] to OMX with shape {self.shape}"
                )
        elif len(value_shape) == 2:
            if value_shape[0] == self.shape[0] and value_shape[1] == self.shape[1]:
                self.add_matrix(key, numpy.asarray(value))
            else:
                raise OMXIncompatibleShape(
                    f"cannot add matrix[{value_shape}] to OMX with shape {self.shape}"
                )
        else:
            raise OMXIncompatibleShape(
                f"cannot add matrix[{value_shape}] which has more than 3 dimnensions to OMX"
            )

    def __getattr__(self, key):
        if key in self.data._v_children:
            if key not in self.lookup._v_children:
                return self.data._v_children[key]
            else:
                raise AttributeError(f"key {key} found in both data and lookup")
        if key in self.lookup._v_children:
            return self.lookup._v_children[key]
        raise AttributeError(f"key {key} not found")

    def get_dataframe(self, matrix, index=None, columns=None):
        """
        Get a matrix array as a pandas.DataFrame.

        This will return the entire content from a single matrix
        array as a pandas.DataFrame, using index or column labels
        from lookup vectors contained in the OMX file, or as
        provided directly.

        Parameters
        ----------
        matrix : str
            The name of the matrix to extract
        index : str or array-like, optional
            The name of the lookup to use as the index, or an array
            of length equal to the first shape dimension.
            If not given, a RangeIndex is used.
        columns : str, optional
            The name of the lookup to use as the columns, or an array
            of length equal to the second shape dimension.
            If not given, a RangeIndex is used.

        Returns
        -------
        pandas.DataFrame
        """
        if isinstance(index, str):
            index = self.lookup._v_children[index]
        if isinstance(columns, str):
            columns = self.lookup._v_children[columns]
        if matrix in self.data._v_children:
            return pandas.DataFrame(
                data=self.data._v_children[matrix],
                index=index,
                columns=columns,
            )
        elif matrix in self.lookup._v_children:
            return pandas.DataFrame(
                data=self.lookup._v_children[matrix],
                index=index,
                columns=[matrix],
            )
        else:
            raise KeyError(matrix)

    def all_matrix_at(self, r, c):
        """
        Get the value from all matrices at a coordinate.

        Parameters
        ----------
        r,c : int
            Coordinate to extract.  This is the zero-based index.

        Returns
        -------
        Dict
            The keys are the names of the matrices, and the value are
            taken from the given coordinates.
        """
        from .util import Dict

        result = Dict()
        for name, vals in self.data._v_children.items():
            result[name] = vals[r, c]
        return result

    def get_rc_dataframe(self, row_indexes, col_indexes, mat_names=None, index=None):
        """
        Build a DataFrame containing values pulled from this OMX.

        Parameters
        ----------
        row_indexes : array-like or int
            The row index within the matrix for each output row.
        col_indexes : array-like or int
            The column index within the matrix for each output row.
            Must have the same shape as `row_indexes`, unless one of
            these is just an integer, in which case that value is
            broadcast to the shape of the other.
        mat_names : Sequence or Mapping, optional
            A sequence of matrix names to draw values from.  Each
            name should be a matrix table that exists in the OMX
            file. If not given, all matrix arrays from the `data`
            node in the OMX file will be used. If given as a mapping,
            the keys are used to identify the columns to draw values
            from, and the mapping is then used to rename the columns.
        index : array-like, or 'rc', optional
            An array to use as the index on the returned DataFrame.
            Set to 'rc' to get a row-and-column MultiIndex.

        Returns
        -------
        pandas.DataFrame
        """
        if mat_names is None:
            mat_names = list(self.data._v_children.keys())
        if isinstance(mat_names, Mapping):
            _mat_names = mat_names.keys()
        elif isinstance(mat_names[0], (tuple, list)):
            _mat_names = [i[0] for i in mat_names]
        else:
            _mat_names = mat_names

        if isinstance(row_indexes, int):
            row_indexes = numpy.full_like(col_indexes, row_indexes)
        elif isinstance(col_indexes, int):
            col_indexes = numpy.full_like(row_indexes, col_indexes)

        data = {mat: self[mat][row_indexes, col_indexes] for mat in _mat_names}

        if index is None:
            try:
                index = row_indexes.index
            except AttributeError:
                pass

        if index is None:
            try:
                index = col_indexes.index
            except AttributeError:
                pass

        if index is None:
            try:
                rows_name = row_indexes.name
            except AttributeError:
                rows_name = "i"
                while rows_name in mat_names:
                    rows_name = rows_name + "_"
            data[rows_name] = row_indexes

            try:
                cols_name = col_indexes.name
            except AttributeError:
                cols_name = "j"
                while cols_name in mat_names:
                    cols_name = cols_name + "_"
            data[cols_name] = col_indexes

        result = pandas.DataFrame.from_dict(data)
        if index is None or (isinstance(index, str) and index == "rc"):
            result = result.set_index([rows_name, cols_name])
        else:
            result.index = index
        if isinstance(mat_names, Mapping):
            result = result.rename(columns=mat_names)
        return result

    def join_rc_dataframe(
        self,
        df,
        rowidx,
        colidx,
        mat_names=None,
        prefix="",
    ):
        """
        Join RC data pulled from this OMX with an existing DataFrame.

        Parameters
        ----------
        df : pandas.DataFrame
            The existing dataframe to join to.
        rowidx, colidx : str or array-like
            The columns to use for the rowindexes and colindexes,
            respectively. Give as a string to name columns in `df`,
            or as an eval-capable instruction, or give an array
            explicitly.
        mat_names : Sequence or Mapping, optional
            A sequence of matrix names to draw values from.  Each
            name should be a matrix table that exists in the OMX
            file. If not given, all matrix arrays from the `data`
            node in the OMX file will be used. If given as a mapping,
            the keys are used to identify the columns to draw values
            from, and the mapping is then used to rename the columns.
        prefix : str, optional
            Add this prefix to every matrix name used.

        Returns
        -------
        pandas.DataFrame
        """
        if isinstance(rowidx, str):
            _row = df.eval(rowidx)
        else:
            _row = rowidx
        if isinstance(colidx, str):
            _col = df.eval(colidx)
        else:
            _col = colidx
        data = self.get_rc_dataframe(
            _row,
            _col,
            mat_names,
            index=df.index,
        )
        if prefix:
            data = data.add_prefix(prefix)
        return df.join(data)

    def import_omx(self, otherfile, tablenames, rowslicer=None, colslicer=None):
        oth = OMX(otherfile, mode="r")
        if tablenames == "*":
            tablenames = oth.data._v_children.keys()
        for tab in tablenames:
            if rowslicer is None and colslicer is None:
                self.add_matrix(tab, oth.data._v_children[tab][:])
            else:
                self.add_matrix(tab, oth.data._v_children[tab][rowslicer, colslicer])

    @classmethod
    def change_cols(
        cls,
        newfile,
        otherfile,
        datanames,
        lookuprownames,
        lookupcolnames,
        newshape,
        newslicecols=None,
        oldslicecols=None,
    ):
        """
        Read an existing OMX into a new file, remapping columns as appropriate.

        The new matrix is filled with blanks, and then the values are set by
        slicing the arrays and assigning the old array to the sliced new.
        (Sorry if this is confusing.)

        Parameters
        ----------
        otherfile
        datanames
        lookuprownames
        lookupcolnames
        newshape
        slicecols
        """
        if oldslicecols is None and newslicecols is None:
            raise TypeError(
                "there is no reason to change cols if you don't slice something"
            )
        if oldslicecols is None:
            oldslicecols = slice(None)
        if newslicecols is None:
            newslicecols = slice(None)

        self = cls(newfile, mode="a")
        self.shape = newshape
        if isinstance(otherfile, str):
            otherfile = OMX(otherfile)
        for name in datanames:
            n = self.add_blank_matrix(
                name,
                atom=otherfile.data._v_children[name].atom,
                shape=newshape,
                complevel=1,
                complib="zlib",
            )
            n[:, newslicecols] = otherfile.data._v_children[name][:, oldslicecols]
        for name in lookuprownames:
            n = self.add_lookup(name, obj=otherfile.lookup._v_children[name][:])
        for name in lookupcolnames:
            n = self.add_blank_lookup(
                name, atom=otherfile.lookup._v_children[name].atom, shape=newshape[1]
            )
            n[newslicecols] = otherfile.lookup._v_children[name][oldslicecols]
        self.flush()
        return self

    # def info(self):
    #     from .model_reporter.art import ART
    #
    #     ## Header
    #     a = ART(
    #         columns=("TABLE", "DTYPE"),
    #         n_head_rows=1,
    #         title="OMX ({},{}) @ {}".format(
    #             self.shape[0], self.shape[1], self.filename
    #         ),
    #         short_title=None,
    #     )
    #     a.addrow_kwd_strings(TABLE="Table", DTYPE="dtype")
    #     ## Content
    #     if len(self.data._v_children):
    #         a.add_blank_row()
    #         a.set_lastrow_iloc(
    #             0, a.encode_cell_value("DATA"), {"class": "parameter_category"}
    #         )
    #     for v_name in sorted(self.data._v_children):
    #         a.addrow_kwd_strings(
    #             TABLE=v_name, DTYPE=self.data._v_children[v_name].dtype
    #         )
    #     if len(self.lookup._v_children):
    #         a.add_blank_row()
    #         a.set_lastrow_iloc(
    #             0, a.encode_cell_value("LOOKUP"), {"class": "parameter_category"}
    #         )
    #     for v_name in sorted(self.lookup._v_children):
    #         a.addrow_kwd_strings(
    #             TABLE=v_name, DTYPE=self.lookup._v_children[v_name].dtype
    #         )
    #     return a

    @classmethod
    def FromDataFrame(cls, frame, *arg, **kwarg):
        """
        Create a new OMX file from a `pandas.DataFrame`.

        Will create a new OMX file containing only lookups (no matrix data).
        The shape of the OMX will be set as square, with size equal to the
        number of rows in the frame.

        Parameters
        ----------
        frame : pandas.DataFrame
            Import this data.
        """
        self = cls(*arg, **kwarg)
        self.shape = (len(frame), len(frame))
        for columnname in frame.columns:
            self.add_lookup(columnname, frame[columnname].values)
        return self

    @classmethod
    def FromCSV(cls, filename, *arg, csv_kwarg=None, **kwarg):
        """
        Create a new OMX file from a csv file.

        This is a convenience function that wraps reading the CSV into a
        `pandas.DataFrame`, and then writing to a new OMX with `FromDataFrame`.

        Parameters
        ----------
        filename : str
            The CSV filename.
        """
        if csv_kwarg is None:
            csv_kwarg = {}
        self = cls.FromDataFrame(pandas.read_csv(filename, **csv_kwarg), *arg, **kwarg)
        return self

    @classmethod
    def FromXLSX(cls, filename, *arg, excel_kwarg=None, **kwarg):
        """
        Create a new OMX file from an excel file.

        This is a convenience function that wraps reading the XLSX into a
        `pandas.DataFrame`, and then writing to a new OMX with `FromDataFrame`.

        Parameters
        ----------
        filename : str
            The CSV filename.
        """
        if excel_kwarg is None:
            excel_kwarg = {}
        if "engine" not in excel_kwarg:
            excel_kwarg["engine"] = "openpyxl"
        self = cls.FromDataFrame(
            pandas.read_excel(filename, **excel_kwarg), *arg, **kwarg
        )
        return self

    def _remake_command(self, cmd, selector=None, receiver=None):
        from tokenize import NAME, OP, tokenize, untokenize

        DOT = (OP, ".")
        COLON = (OP, ":")
        COMMA = (OP, ",")
        OBRAC = (OP, "[")
        CBRAC = (OP, "]")
        from io import BytesIO

        recommand = []

        if receiver:
            recommand += [
                (NAME, receiver),
                OBRAC,
                COLON,
                CBRAC,
                (OP, "="),
            ]

        try:
            cmd_encode = cmd.encode("utf-8")
        except AttributeError:
            cmd_encode = str(cmd).encode("utf-8")
        g = tokenize(BytesIO(cmd_encode).readline)
        if selector is None:
            screen_tokens = [
                COLON,
            ]
        else:
            screen_tokens = [
                (NAME, "selector"),
            ]
        for toknum, tokval, _, _, _ in g:
            if toknum == NAME and tokval in self.data:
                # replace NAME tokens
                partial = [
                    (NAME, "self"),
                    DOT,
                    (NAME, "data"),
                    DOT,
                    (NAME, tokval),
                    OBRAC,
                ]
                partial += screen_tokens
                if len(self._groupnode._v_children[tokval].shape) > 1:
                    partial += [
                        COMMA,
                        COLON,
                    ]
                if len(self._groupnode._v_children[tokval].shape) > 2:
                    partial += [
                        COMMA,
                        COLON,
                    ]
                if len(self._groupnode._v_children[tokval].shape) > 3:
                    partial += [
                        COMMA,
                        COLON,
                    ]
                partial += [
                    CBRAC,
                ]
                recommand.extend(partial)
            elif toknum == NAME and tokval in self.lookup:
                # replace NAME tokens
                partial = [
                    (NAME, "self"),
                    DOT,
                    (NAME, "lookup"),
                    DOT,
                    (NAME, tokval),
                    OBRAC,
                ]
                partial += screen_tokens
                if len(self._groupnode._v_children[tokval].shape) > 1:
                    partial += [
                        COMMA,
                        COLON,
                    ]
                if len(self._groupnode._v_children[tokval].shape) > 2:
                    partial += [
                        COMMA,
                        COLON,
                    ]
                if len(self._groupnode._v_children[tokval].shape) > 3:
                    partial += [
                        COMMA,
                        COLON,
                    ]
                partial += [
                    CBRAC,
                ]
                recommand.extend(partial)
            else:
                recommand.append((toknum, tokval))
        # print("<recommand>")
        # print(recommand)
        # print("</recommand>")
        ret = untokenize(recommand).decode("utf-8")
        from .util.aster import asterize

        return asterize(ret, mode="exec" if receiver is not None else "eval"), ret

    def _evaluate_single_item(self, cmd, selector=None, receiver=None):
        j, j_plain = self._remake_command(
            cmd,
            selector=selector,
            receiver="receiver" if receiver is not None else None,
        )
        # important globals

        try:
            if receiver is not None:
                exec(j)
            else:
                return eval(j)
        except Exception as exc:
            args = exc.args
            if not args:
                arg0 = ""
            else:
                arg0 = args[0]
            arg0 = arg0 + f'\nwithin parsed command: "{cmd!s}"'
            arg0 = arg0 + f'\nwithin re-parsed command: "{j_plain!s}"'
            if selector is not None:
                arg0 = arg0 + f'\nwith selector: "{selector!s}"'
            if "max" in cmd:
                arg0 = (
                    arg0
                    + '\n(note to get the maximum of arrays use "fmax" not "max")'.format()
                )
            if "min" in cmd:
                arg0 = (
                    arg0
                    + '\n(note to get the minimum of arrays use "fmin" not "min")'.format()
                )
            if isinstance(exc, NameError):
                badname = str(exc).split("'")[1]
                goodnames = dir()
                from .util.text_manip import case_insensitive_close_matches

                did_you_mean_list = case_insensitive_close_matches(
                    badname, goodnames, n=3, cutoff=0.1, excpt=None
                )
                if len(did_you_mean_list) > 0:
                    arg0 = (
                        arg0
                        + "\n"
                        + "did you mean {}?".format(
                            " or ".join(f"'{s}'" for s in did_you_mean_list)
                        )
                    )
            exc.args = (arg0,) + args[1:]
            raise

    def to_dataset(self, zone_id=None, dim_names=("otaz", "dtaz")):
        import numpy as np
        import xarray as xr

        z = xr.Dataset()
        if zone_id is None:
            z = z.assign_coords(
                {
                    dim_names[0]: np.arange(1, self.shape[0] + 1),
                    dim_names[1]: np.arange(1, self.shape[1] + 1),
                }
            )
        else:
            z = z.assign_coords(
                {
                    dim_names[0]: self[zone_id],
                    dim_names[1]: self[zone_id],
                }
            )
        for k in self.data._v_children:
            z = z.assign({k: ((dim_names[0], dim_names[1]), self[k])})
        for k in self.lookup._v_children:
            if k == zone_id:
                continue
            k_ = self[k][:]
            if k_.size == self.shape[0]:
                z = z.assign({k: (dim_names[0], self[k])})
            elif k_.size == self.shape[1]:
                z = z.assign({k: (dim_names[1], self[k])})
            else:
                raise OMXIncompatibleShape
        return z



from .._optional import jax, jnp


def lhs_draw_maker(n_draws, n_cols):
    """
    Create a jitted function that makes draws.

    Parameters
    ----------
    n_draws, n_cols : int

    Returns
    -------
    Callable
        Accepts a jax random state key, returns a new key and the draws
    """

    @jax.jit
    def lhs_draw_column(carry, x):
        key = carry
        k1, k2, k3 = jax.random.split(key, 3)
        return k3, jax.random.permutation(
            k1,
            (jnp.arange(n_draws) + jax.random.uniform(k2, shape=(n_draws,))) / n_draws,
        )

    @jax.jit
    def lhs_draws(key):
        return jax.lax.scan(lhs_draw_column, key, None, length=n_cols)

    return lhs_draws



import contextlib
import warnings


@contextlib.contextmanager
def ignore_warnings():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        yield


class NoDefaultValueWarning(Warning):
    pass


def no_default_value(message):
    # warnings.warn(message, NoDefaultValueWarning, stacklevel=2)
    pass



def example(extract='m', estimate=False):
    import numpy as np
    import pandas as pd

    import larch as lx
    from larch import P, X

    hh, pp, tour, skims, emp = lx.example(200, ["hh", "pp", "tour", "skims", "emp"])

    hh["INCOME_GRP"] = pd.qcut(hh.INCOME, 3)

    logsums_file = lx.example(202, output_file="logsums.zarr")
    logsums = lx.DataArray.from_zarr("logsums.zarr")

    ca = lx.Dataset.construct(
        {"logsum": logsums},
        caseid="TOURID",
        alts=skims.TAZ_ID,
    )
    ca

    emp.info()

    tree = lx.DataTree(
        base=ca,
        tour=tour.rename_axis(index="TOUR_ID"),
        hh=hh.set_index("HHID"),
        person=pp.set_index("PERSONID"),
        emp=emp,
        skims=lx.Dataset.construct.from_omx(skims),
        relationships=(
            "base.TAZ_ID @ emp.TAZ",
            "base.TOURID -> tour.TOUR_ID",
            "tour.HHID @ hh.HHID",
            "tour.PERSONID @ person.PERSONID",
            "hh.HOMETAZ @ skims.otaz",
            "base.TAZ_ID @ skims.dtaz",
        ),
    )

    m = lx.Model(datatree=tree)
    m.title = "Exampville Work Tour Destination Choice v1"

    m.quantity_ca = (
        +P.EmpRetail_HighInc * X("RETAIL_EMP * (INCOME>50000)")
        + P.EmpNonRetail_HighInc * X("NONRETAIL_EMP") * X("INCOME>50000")
        + P.EmpRetail_LowInc * X("RETAIL_EMP") * X("INCOME<=50000")
        + P.EmpNonRetail_LowInc * X("NONRETAIL_EMP") * X("INCOME<=50000")
    )

    m.quantity_scale = P.Theta

    m.utility_ca = +P.logsum * X.logsum + P.distance * X.AUTO_DIST

    m.choice_co_code = "tour.DTAZ"

    m.plock(EmpRetail_HighInc=0, EmpRetail_LowInc=0)

    mj = m.copy()

    m.compute_engine = "numba"

    m.loglike()

    m.d_loglike()

    mj.d_loglike()
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True)

    resultj = mj.maximize_loglike(stderr=True)



    m.histogram_on_idca_variable("AUTO_DIST")

    m.histogram_on_idca_variable("RETAIL_EMP")

    m.histogram_on_idca_variable("AUTO_DIST", bins=40, span=(0, 10))

    m.histogram_on_idca_variable(
        "AUTO_DIST",
        x_label="Distance (miles)",
        bins=26,
        span=(0, 13),
        filter_co="INCOME<10000",
    )

    report = lx.Reporter(title=m.title)

    report << "# Parameter Summary" << m.parameter_summary()

    report << "# Parameter Summary" << m.parameter_summary()

    report << "# Estimation Statistics" << m.estimation_statistics()

    report << "# Utility Functions" << m.utility_functions()

    figure = m.histogram_on_idca_variable(
        "AUTO_DIST",
        bins=30,
        span=(0, 15),
        x_label="Distance (miles)",
    )
    report << "# Visualization"
    report << figure

    report.save(
        "exampville_dest_choice.html",
        overwrite=True,
        metadata=m.dumps(),
    )
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import importlib
import os.path
from collections.abc import Mapping

import numpy as np
from addicty import Dict
from yaml import SafeDumper

from .constraints import ParametricConstraintList
from .linear import DictOfAlts, DictOfLinearFunction, LinearComponent, LinearFunction
from .tree import NestingTree

SafeDumper.add_representer(
    np.float64, lambda dumper, data: dumper.represent_float(float(data))
)
SafeDumper.add_representer(
    np.float32, lambda dumper, data: dumper.represent_float(float(data))
)
SafeDumper.add_representer(
    np.int32, lambda dumper, data: dumper.represent_int(int(data))
)
SafeDumper.add_representer(
    np.int64, lambda dumper, data: dumper.represent_int(int(data))
)
SafeDumper.add_representer(
    np.bool_, lambda dumper, data: dumper.represent_bool(bool(data))
)
SafeDumper.add_representer(
    np.str_, lambda dumper, data: dumper.represent_str(str(data))
)
SafeDumper.add_representer(
    np.ndarray, lambda dumper, data: dumper.represent_list(list(data))
)
SafeDumper.add_representer(
    LinearComponent, lambda dumper, data: dumper.represent_dict(data.to_dict())
)
SafeDumper.add_representer(
    LinearFunction, lambda dumper, data: dumper.represent_list(list(data))
)
SafeDumper.add_representer(
    DictOfLinearFunction, lambda dumper, data: dumper.represent_dict(dict(data))
)
SafeDumper.add_representer(
    DictOfAlts, lambda dumper, data: dumper.represent_dict(dict(data))
)
SafeDumper.add_representer(
    ParametricConstraintList, lambda dumper, data: dumper.represent_list(list(data))
)
SafeDumper.add_representer(
    ParametricConstraintList, lambda dumper, data: dumper.represent_list(list(data))
)
SafeDumper.add_representer(
    NestingTree, lambda dumper, data: dumper.represent_dict(data.to_dict())
)


def save_model(m, filename=None, format="yaml", overwrite=False):
    if format == "yaml":
        if overwrite and os.path.isfile(filename):
            os.unlink(filename)
        return _save_model_yaml(m, filename)
    elif format == "raw":
        return _save_model_yaml(m, None, raw=True)
    else:
        raise NotImplementedError(f"saving in {format} format")


def load_model(filename_or_content):
    if isinstance(filename_or_content, str) and os.path.exists(filename_or_content):
        if filename_or_content.endswith(".html") or filename_or_content.endswith(
            ".xhtml"
        ):
            from .. import read_metadata

            try:
                y = read_metadata(filename_or_content)
            except Exception:
                raise
            else:
                if isinstance(y, str) and "\n" in y:
                    y = Dict.load(y)
                if y is not None:
                    filename_or_content = y
    return _load_model_yaml(filename_or_content)


def _save_model_yaml(m, filename, skipping=(), raw=False):
    x = Dict()

    def savethis(attr, transformer=None, wrapper=None):
        if attr not in skipping:
            i = getattr(m, attr, None)
            if i:
                try:
                    if transformer is not None:
                        j = getattr(i, transformer)
                        if callable(j):
                            i = j()
                        else:
                            i = j
                    elif wrapper is not None:
                        i = wrapper(i)
                except AttributeError:
                    pass
                else:
                    x[attr] = i

    x.model_type = f"{m.__class__.__module__}.{m.__class__.__name__}"
    savethis("availability_any")
    savethis("availability_ca_var")
    savethis("availability_co_vars", "to_dict")
    savethis("choice_any")
    savethis("choice_ca_var")
    savethis("choice_co_code")
    savethis("choice_co_vars", "to_dict")
    savethis("common_draws")
    savethis("compute_engine")
    savethis("constraint_intensity")
    savethis("constraint_sharpness")
    savethis("constraints")
    savethis("float_dtype", "__name__")
    savethis("graph", "to_dict")
    savethis("groupid")
    savethis("index_name")
    savethis("logsum_parameter")
    savethis("mixtures", "to_list")
    savethis("n_draws")
    savethis("parameters", "to_dict")
    savethis("prerolled_draws")
    savethis("quantity_ca")
    savethis("quantity_scale")
    savethis("title")
    savethis("utility_ca")
    savethis("utility_co")
    savethis("weight_co_var")
    savethis("weight_normalization")

    _models = getattr(m, "_models", None)
    if _models is not None:
        submodels = Dict()
        for k, v in _models.items():
            submodels[k] = _save_model_yaml(v, None, skipping=("parameters"), raw=True)
        x["_models"] = submodels

    if raw:
        return x
    return x.dump(filename)


def _load_model_yaml(filename_or_content):
    if isinstance(filename_or_content, str) and os.path.exists(filename_or_content):
        content = Dict.load(filename_or_content)
    elif isinstance(filename_or_content, Mapping):
        content = filename_or_content
    else:
        raise TypeError("filename_or_content must be an existing file or a Dict")

    model_type = content.get("model_type").split(".")
    model_module = ".".join(model_type[:-1])
    module = importlib.import_module(model_module)
    cls = getattr(module, model_type[-1])
    return cls.from_dict(content)



# This file is under the Creative Commons Attribution Share Alike license
# https://creativecommons.org/licenses/by-sa/3.0/
# original source:
# https://stackoverflow.com/questions/20643184/using-python-threads-to-make-thousands-of-calls-to-a-slow-api-with-a-rate-limit

import functools
import time
from collections.abc import Iterator
from threading import Lock


class BlockingRateLimiter(Iterator):
    """Iterator that yields a value at most once every 'interval' seconds."""

    def __init__(self, interval):
        self.lock = Lock()
        self.interval = interval
        self.next_yield = 0

    def __next__(self):
        with self.lock:
            t = time.monotonic()
            if t < self.next_yield:
                time.sleep(self.next_yield - t)
                t = time.monotonic()
            self.next_yield = t + self.interval


class NonBlockingRateLimiter:
    def __init__(self, interval):
        self.lock = Lock()
        self.interval = interval
        self.next_greenlight = 0

    def __call__(self, fn):
        @functools.wraps(fn)
        def decorated(*args, **kwargs):
            t = time.monotonic()
            if t >= self.next_greenlight:
                with self.lock:
                    self.next_greenlight = t + self.interval
                    fn(*args, **kwargs)

        return decorated

    def __bool__(self):
        t = time.monotonic()
        if t >= self.next_greenlight:
            with self.lock:
                self.next_greenlight = t + self.interval
                return True
        return False


_global_rate_limiters = {}


def GlobalRateLimiter(tag, interval=1, wait_now=True):
    global _global_rate_limiters
    if tag not in _global_rate_limiters:
        _global_rate_limiters[tag] = NonBlockingRateLimiter(interval)
    if wait_now:
        return next(_global_rate_limiters[tag])



from ... import DataFrames, Model
from .. import Dict
from .general import (
    apply_coefficients,
    dict_of_linear_utility_from_spec,
    remove_apostrophes,
    simple_simulate_data,
)


def auto_ownership_model(
    name="auto_ownership",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    data = simple_simulate_data(
        name=name,
        edb_directory=edb_directory,
        values_index_col="household_id",
    )
    coefficients = data.coefficients
    # coef_template = data.coef_template # not used
    spec = data.spec
    chooser_data = data.chooser_data
    _settings = data.settings

    altnames = list(spec.columns[3:])
    altcodes = range(len(altnames))

    chooser_data = remove_apostrophes(chooser_data)
    chooser_data.fillna(0, inplace=True)

    # Remove choosers with invalid observed choice
    chooser_data = chooser_data[chooser_data["override_choice"] >= 0]

    m = Model()
    # One of the alternatives is coded as 0, so
    # we need to explicitly initialize the MNL nesting graph
    # and set to root_id to a value other than zero.
    m.initialize_graph(alternative_codes=altcodes, root_id=99)

    m.utility_co = dict_of_linear_utility_from_spec(
        spec,
        "Label",
        dict(zip(altnames, altcodes)),
    )

    apply_coefficients(coefficients, m)

    d = DataFrames(
        co=chooser_data,
        av=True,
        alt_codes=altcodes,
        alt_names=altnames,
    )

    m.dataservice = d
    m.choice_co_code = "override_choice"

    if return_data:
        return m, Dict(
            edb_directory=data.edb_directory,
            chooser_data=chooser_data,
            coefficients=coefficients,
            spec=spec,
            altnames=altnames,
            altcodes=altcodes,
        )

    return m
